// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Aug 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  ai.js
//    Description: Javascript that handles general ai related functions
//       for the game
// ============================================================================

/**
* NAMESPACE: Game.ai
* USE: Handles ai related functions
*/

// Declare the namespace
registerNamespace('Game.ai');

// Declare the namespace members
(function(namespace) {
    'use strict';

// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;

    // Ai difficulty enum
    exposed.difficulty = Object.freeze({easy:0, normal:1, hard:2, impossible:3});

    /**
    * FUNCTION: getMove
    * USE: Gets a difficulty move for the given player
    * @param playerId: The player to get the move for
    * @param difficulty: The type of difficulty move to make
    * @return: The game piece id of the move to make
    */
    exposed.getMove = (playerId, difficulty = null) => {
        let move = getDifficultyMove(playerId, difficulty);
        return move;
    }

    /**
    * FUNCTION: getDescription
    * USE: Gets the string description of the difficulty enum
    * @param difficulty: The difficulty type
    * @return: The enum string description
    */
    exposed.getDescription = (difficulty) => {
        let description = '';
        switch (difficulty) {
            case exposed.difficulty.easy:
                description = 'Easy';
                break;
            case exposed.difficulty.normal:
                description = 'Normal';
                break;
            case exposed.difficulty.hard:
                description = 'Hard';
                break;
            case exposed.difficulty.impossible:
                description = 'Impossible';
                break;
        }
        return description.trim();
    }

// -- Private data --
    /**
    * FUNCTION: getDifficultyMove
    * USE: Gets a difficulty move the player can make
    * @param playerId: The player to get the move for
    * @param difficulty: The type of difficulty move to make
    * @return: The game piece id of the move to make
    */
    let getDifficultyMove = (playerId, difficulty) => {
        // Make sure parameters are valid
        difficulty = difficulty != null ? difficulty : exposed.difficulty.normal;

        let state = createGameState(playerId);
        let available = state.emptyCells();
        let move = null;

        // Get a move for the selected difficulty
        if (available.length > 0) {
            let action = null;
            // If this is the first move, get a random action to reduce ai predictability
            if (state.board.length === available.length) {
                action = getRandomMove(state, available);
            } else {
                // Get a move based on the difficulty
                switch (difficulty) {
                    case exposed.difficulty.easy:
                        action = getEasyMove(state, available);
                        break;
                    case exposed.difficulty.normal:
                        action = getNormalMove(state, available);
                        break;
                    case exposed.difficulty.hard:
                        action = getHardMove(state, available);
                        break;
                    default:
                        action = getImpossibleMove(state, available);
                        break;
                }
            }
            move = state.getPieceId(action.movePosition);
        }
        return move;
    }

    /**
    * FUNCTION: createGameState
    * USE: Gets a State object that represents the current state of the game
    * @param playerId: The player to get the move for
    * @return: A State object of the current state of the game
    */
    let createGameState = (playerId) => {
        let matchInfo = getMatchInfo();
        let currentState = new State();
        currentState.maximizingPlayer = playerId;
        currentState.players = matchInfo.players;
        currentState.board = matchInfo.board;
        currentState.turn = matchInfo.players.indexOf(playerId);
        // Make sure the players turn is found
        if (currentState.turn === -1) {
            throw new Error('Unable to determine the maximizing players turn');
        }
        return currentState;
    }

    /**
    * FUNCTION: getMatchInfo
    * USE: Gets information about the current match
    * @return: Information about the current match
    */
    let getMatchInfo = () => {
        let board = [];
        let match = Game.status.getCurrentMatch();
        let gameMoves = match.getMoves();
        let players = match.getPlayers();

        // Go through the rows to get the pieces
        let gamePieces = Game.board.getPieces();
        for (let indexCol = 0; indexCol < gamePieces.length; ++indexCol) {
            let gamePiece = gamePieces[indexCol];
            let pieceId = Game.board.getPieceId(gamePiece);
            let moveInfo = isPieceSelected(gameMoves, pieceId);

            // Mark the index with the selected player
            board[indexCol] = {
                playerId: moveInfo.isSelected ? moveInfo.playerId : null,
                pieceId: pieceId,
            };
        }
        return {
            board: board,
            players: players,
        };
    }

    /**
    * FUNCTION: isPieceSelected
    * USE: Determines if the piece id is selected and returns the player that selected it
    * @param gameMoves: The current player moves in the current match
    * @param pieceId: The game piece id to check for
    * @return: Information determining if the piece has been selected
    */
    let isPieceSelected = (gameMoves, pieceId) => {
        let playerId = null;
        let isSelected = false;
        for (let key in gameMoves) {
            let playerMoves = gameMoves[key];
            if (playerMoves.indexOf(pieceId) > -1) {
                playerId = key;
                isSelected = true;
                break;
            }
        }
        return {
            isSelected: isSelected,
            playerId: playerId,
        };
    }

    /**
    * FUNCTION: getEasyMove
    * USE: Gets an easy move for the maximizing player,
    *   that is: mix between randomly chosen decisions
    * @param state: The current state of the game
    * @param available: The current moves that are available
    * @return: The action that the maximizing player should make
    */
    let getEasyMove = (state, available) => {
        let action = null;
        let percentage = randomFromTo(1, 100);
        if (percentage >= 40) {
            action = getRandomMove(state, available);
        } else {
            action = getNormalMove(state, available);
        }
        return action;
    }

    /**
    * FUNCTION: getNormalMove
    * USE: Gets a normal move for the maximizing player,
    *   that is: mix between choosing suboptimal minimax decisions
    * @param state: The current state of the game
    * @param available: The current moves that are available
    * @return: The action that the maximizing player should make
    */
    let getNormalMove = (state, available = null) => {
        // Get the best available moves
        let availableActions = getBestMoves(state, available);

        // First take the winning action. If none exists, take the
        // optimal action a percentage of the time, and take a suboptimal action otherwise
        let chosenAction = findWinningAction(availableActions);
        if (!chosenAction) {
            // Default to the optimal action
            chosenAction = availableActions[0];
            let percentage = randomFromTo(1, 100);
            if (percentage >= 50) {
                // Choose a random action
                chosenAction = getRandomArrayItem(availableActions);
            } else if (percentage >= 25) {
                if (availableActions.length >= 2) {
                    // If there is two or more available actions, choose the 1st suboptimal
                    chosenAction = availableActions[1];
                }
            }
        }
        return chosenAction;
    }

    /**
    * FUNCTION: getHardMove
    * USE: Gets an hard move for the maximizing player,
    *   that is: choose a mix between optimal and suboptimal minimax decision
    * @param state: The current state of the game
    * @param available: The current moves that are available
    * @return: The action that the maximizing player should make
    */
    let getHardMove = (state, available = null) => {
        // Get the best available moves
        let availableActions = getBestMoves(state, available);

        // Default to the optimal action
        let chosenAction = availableActions[0];

        // When half the board is filled up, if there is x or more available actions,
        // choose a random optimal choice
        if (availableActions.length <= Math.floor(state.board.length / 2)) {
            let firstXCount = Math.min(2, availableActions.length);
            let firstXActions = availableActions.slice(0, firstXCount);

            // If there is a winning action, choose that one
            let winningAction = findWinningAction(firstXActions);
            if (winningAction) {
                chosenAction = winningAction;
            } else {
                chosenAction = getRandomArrayItem(firstXActions);
            }
        }
        return chosenAction;
    }

    /**
    * FUNCTION: getImpossibleMove
    * USE: Gets an impossible move for the maximizing player,
    *   that is: choose the optimal minimax decision
    * @param state: The current state of the game
    * @param available: The current moves that are available
    * @return: The action that the maximizing player should make
    */
    let getImpossibleMove = (state, available = null) => {
        // Get the best available moves
        let availableActions = getBestMoves(state, available);

        // Default to the optimal action
        let chosenAction = availableActions[0];
        return chosenAction;
    }

    /**
    * FUNCTION: getRandomMove
    * USE: Gets a random move for the maximizing player
    * @param state: The current state of the game
    * @param available: The current moves that are available
    * @return: The action that the maximizing player should make
    */
    let getRandomMove = (state, available = null) => {
        available = available || state.emptyCells();
        let randomCell = getRandomArrayItem(available);
        let action = new Action(randomCell);
        return action;
    }

    /**
    * FUNCTION: getBestMoves
    * USE: Gets the best moves in descending order for the maximizing player
    * @param state: The current state of the game
    * @param available: The current moves that are available
    * @return: The action that the maximizing player should make
    */
    let getBestMoves = (state, available = null) => {
        available = available || state.emptyCells();
        // Get and calculate the score for each available actions to the maximizing player
        let availableActions = available.map((pos) => {
            // Create the action object
            let action = new Action(pos);
            // Get next state by applying the action
            let nextState = action.applyTo(state);
            // Calculate and set the actions minimax value
            action.minimaxVal = minimaxValue(nextState);
            return action;
        });

        // Sort the actions list by score in descending order with maximum minimax at first
        availableActions.sort(Action.DESCENDING);
        return availableActions;
    }

    /**
    * FUNCTION: findWinningAction
    * USE: Gets the first winning action from a list of actions (if any)
    * @param actions: A list of actions the player could make
    * @return: The winning action (if any)
    */
    let findWinningAction = (actions) => {
        return actions.find((action) => action.minimaxVal > 0);
    }

    /**
    * FUNCTION: minimaxValue
    * USE: Computes the minimax value of a game state
    * @param state: The state to calculate its minimax value
    * @return: The minimax value of the state
    */
    let minimaxValue = (state) => {
        // Minimax value to compute
        let stateScore = null;
        if (state.isTerminal()) {
            // A terminal game state is the base case
            stateScore = state.score();
        } else {
            if (state.getCurrentPlayer() == state.maximizingPlayer) {
                // Maximizing player: initialize to a value smaller than any possible score
                stateScore = -1987;
            } else {
                // Minimizing player: initialize to a value larger than any possible score
                stateScore = 1991;
            }

            // Get the next available states using the info form available positions
            let availablePositions = state.emptyCells();
            for (let pos of availablePositions) {
                // Get the next state
                let nextState = new Action(pos).applyTo(state);

                // Calculate the minimax value for all available next states
                // and evaluate the current state's value
                let nextScore = minimaxValue(nextState);
                if (state.getCurrentPlayer() == state.maximizingPlayer) {
                    // Maximizing player: update stateScore if nextScore is larger
                    if (nextScore > stateScore) {
                        stateScore = nextScore;
                    }
                } else {
                    // Minimizing player: update stateScore if nextScore is smaller
                    if (nextScore < stateScore) {
                        stateScore = nextScore;
                    }
                }
            }
        }
        return stateScore;
    }

    /**
    * Class: State
    * USE: Represents a state in the game
    * @param old: Old state to intialize the new state
    */
    class State {
        constructor(old) {
            this.turn = null;
            this.players = null;
            this.maximizingPlayer = null;
            this.winner = null;
            this.isTie = false;
            this.minimizingMovesCount = 0;
            this.board = [];

            // State is constructed using a copy of another state
            if (typeof old !== 'undefined') {
                // Make a copy of the board from the old state
                for (let indexCol = 0; indexCol < old.board.length; ++indexCol) {
                    this.board.push(Object.assign({}, old.board[indexCol]));
                }
                this.minimizingMovesCount = old.minimizingMovesCount;
                this.turn = old.turn;
                this.players = old.players;
                this.maximizingPlayer = old.maximizingPlayer;
                this.winner = old.winner;
                this.isTie = old.isTie;
            }
        }

        /**
        * FUNCTION: advanceTurn
        * USE: Advances the turn to the next player in the state
        * @return: N/A
        */
        advanceTurn() {
            this.turn = wrap(this.turn + 1, this.players.length);
        }

        /**
        * FUNCTION: emptyCells
        * USE: Gets the empty cells in the state
        * @return: A list of indices of all empty cells
        */
        emptyCells() {
            let cells = [];
            for (let indexCol = 0; indexCol < this.board.length; ++indexCol) {
                if (this.board[indexCol].playerId == null) {
                    cells.push(indexCol);
                }
            }
            return cells;
        }

        /**
        * FUNCTION: getCurrentPlayer
        * USE: Gets the current turn player
        * @return: The current turn player
        */
        getCurrentPlayer() {
            return this.players[this.turn];
        }

        /**
        * FUNCTION: getPieceId
        * USE: Gets the gamepiece id at the given board index position
        * @param pos: The board index position
        * @return: The gamepiece id at the given board index
        */
        getPieceId(pos) {
            return this.board[pos].pieceId;
        }

        /**
        * FUNCTION: isTerminal
        * USE: Checks if the state is a terminal state or not
        *   The state result is updated to reflect the result of the game
        * @return: True if it's terminal, false otherwise
        */
        isTerminal() {
            let board = this.board;
            let players = this.players;
            let playerMoves = {};
            let isTie = true;

            // Get the moves for the player
            for (let indexCol = 0; indexCol < board.length; ++indexCol) {
                // Check if cell is empty
                if (board[indexCol].playerId != null) {
                    let playerId = board[indexCol].playerId;
                    let pieceId = board[indexCol].pieceId;

                    // Create a list of player moves
                    if (!playerMoves.hasOwnProperty(playerId)) {
                        playerMoves[playerId] = [];
                    }
                    // Set the player move to the piece id
                    playerMoves[playerId].push(pieceId);
                } else {
                    // Theres still open moves available
                    isTie = false;
                }
            }
            // Check for a winner
            for (let playerId of players) {
                let moves = playerMoves[playerId];
                if (!moves || moves.length < 1) {
                    continue;
                }
                let checkWinResult = Game.status.checkWin(moves);
                if (checkWinResult.winningComboFound) {
                    // Set the winning player
                    this.winner = playerId;
                    return true;
                }
            }
            if (isTie) {
                // The game is draw
                this.isTie = true;
                return true;
            }
            return false;
        }

        /**
        * FUNCTION: score
        * USE: Calculates the score of the maximizing player in a given terminal state
        * @return: The score calculated for the maximizing player
        */
        score() {
            let stateScore = null;
            if (this.isTie) {
                // Its a draw
                stateScore = 0;
            } else {
                let baseScore = this.board.length + 1;
                if (this.winner == this.maximizingPlayer) {
                    // The maximizing player won
                    stateScore = baseScore - this.minimizingMovesCount;
                } else {
                    // The maximizing player lost
                    stateScore = (baseScore * -1) + this.minimizingMovesCount;
                }
            }
            return stateScore;
        }
    }

    /**
    * Class: Action
    * USE: Constructs an action that the player could make
    * @param pos: The board index position the player should make
    */
    class Action {
        constructor(pos) {
            // The position on the board that the action would put the player on
            this.movePosition = pos;
            // The minimax value of the state that the action leads to when applied
            this.minimaxVal = 0;
        }

        /**
        * FUNCTION: applyTo
        * USE: Applies the action to a state to get the next state
        * @param state: The state to apply the action to
        * @return: The next state after applying the action
        */
        applyTo(state) {
            // Create the next state
            let next = new State(state);

            // Put the player on the board
            next.board[this.movePosition].playerId = next.getCurrentPlayer();

            if (next.getCurrentPlayer() != next.maximizingPlayer) {
                ++next.minimizingMovesCount;
            }
            next.advanceTurn();
            return next;
        }

        /**
        * FUNCTION: ASCENDING
        * USE: Defines a rule for sorting Actions in ascending manner
        * @param firstAction: The first action in a pairwise sort
        * @param secondAction: The second action in a pairwise sort
        * @return: -1, 1, or 0
        */
        static ASCENDING(firstAction, secondAction) {
            if (firstAction.minimaxVal < secondAction.minimaxVal)
                return -1; //indicates that firstAction goes before secondAction
            else if(firstAction.minimaxVal > secondAction.minimaxVal)
                return 1; //indicates that secondAction goes before firstAction
            else
                return 0; //indicates a tie
        }

        /**
        * FUNCTION: DESCENDING
        * USE: Defines a rule for sorting AIActions in descending manner
        * @param firstAction: The first action in a pairwise sort
        * @param secondAction: The second action in a pairwise sort
        * @return: -1, 1, or 0
        */
        static DESCENDING(firstAction, secondAction) {
            if(firstAction.minimaxVal > secondAction.minimaxVal)
                return -1; //indicates that firstAction goes before secondAction
            else if(firstAction.minimaxVal < secondAction.minimaxVal)
                return 1; //indicates that secondAction goes before firstAction
            else
                return 0; //indicates a tie
        }
    }

    /**
    * FUNCTION: wrap
    * USE: Performs a circular wrap of the given values
    * @param left: The lowerbound value
    * @param right: The upperbound value
    * @return: The circular wrapped value
    */
    let wrap = (left, right) => {
        return ((left % right) + right) % right;
    }

    /**
    * FUNCTION: getRandomArrayItem
    * USE: Gets a random item in the array
    * @param arry: The array to get a value from
    * @return: A random array item
    */
    let getRandomArrayItem = (arry) => {
        return arry[Math.floor(Math.random() * arry.length)];
    }

    /**
    * FUNCTION: randomFromTo
    * USE: Gets a random number in a range
    * @param from: The lowerbound number
    * @param to: The upperbound number
    * @return: A random number in the range
    */
    let randomFromTo = (from, to) => {
        return Math.floor(Math.random() * (to - from + 1) + from);
    }
}(Game.ai)); // http://programmingnotes.org/