// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  board.js
//    Description: Javascript that handles gameboard related operations
// ============================================================================ 

/**
* NAMESPACE: Game.board
* USE: Handles gameboard related operations.
*/

// Declare the namespace
registerNamespace('Game.board');

// Declare the namespace members
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
    
    // Set class names and other shared data
    const settings = {	
        classNameGameboard: '.gameboard',
        classNameGameboardRow: '.gameboard-row',
        classNameGamepiece: '.gamepiece',
        classNameGamepieceText: '.gamepiece-text',
        classNameDisabled: '.disabled',
        classNameHighlight: '.highlight',
        dataNameGamePieceId: 'data-id',
        cleanClassName: function(str) {
            return str ? str.trim().replace('.', '') : '';
        },
        stylesApplied: [],
    };
    exposed.settings = settings;
    
    /**
    * FUNCTION: get
    * USE: Returns the gameboard descendant of the object on
    * 	which this method was called.
    * @param element: JavaScript element to search for the gameboard
    * @return: The gameboard descendant found in the given element.
    */	
    exposed.get = function(element = document) {
        return element.querySelector(settings.classNameGameboard);
    }
    
    /**
    * FUNCTION: getRows
    * USE: Returns gameboard rows that are descendants of the object on
    * 	which this method was called.
    * @param element: JavaScript element to search for gameboard rows
    * @return: Gameboard row descendants found in the given element.
    */	
    exposed.getRows = function(element = document) {
        return element.querySelectorAll(settings.classNameGameboardRow);
    }

    /**
    * FUNCTION: getPieces
    * USE: Returns gameboard pieces that are descendants of the object on
    * 	which this method was called.
    * @param element: JavaScript element to search for gameboard pieces
    * @return: Gameboard row descendants found in the given element.
    */	
    exposed.getPieces = function(element = document) {
        return element.querySelectorAll(settings.classNameGamepiece);
    }
    
    /**
    * FUNCTION: getPieceText
    * USE: Returns gamepiece text descendant of the object on
    * 	which this method was called.
    * @param element: JavaScript element to search for gamepiece text
    * @return: Gamepiece text descendant found in the given element.
    */	
    exposed.getPieceText = function(element = document) {
        return element.querySelector(settings.classNameGamepieceText);
    }
    
    /**
    * FUNCTION: clear
    * USE: Clears the gameboard and returns it to its default state
    * @param element: JavaScript element to search for the gameboard 
    * 	and gamepieces.	
    * @return: N/A.
    */		
    exposed.clear = function(element = document) {
        // Find the gameboard and make sure it exists
        let gameboard = exposed.get(element);
        if (!gameboard) {
            throw new Error('Unable to find the gameboard!');
        }
        
        // Find gameboard rows and make sure they exists
        let gameboardRows = exposed.getRows(gameboard);	
        if (!gameboardRows) {
            throw new Error('Unable to find gameboard rows!');
        }

        // Remove styles from the gameboard
        removeClassApplied(gameboard);

        // Go through each row and reset data
        for (const gameboardRow of gameboardRows) {
            // Remove styles from the gameboard row
            removeClassApplied(gameboardRow);

            // Get gamepieces for the row and make sure they exist
            let gamePieces = exposed.getPieces(gameboardRow);
            if (!gamePieces) {
                continue;
            }
            
            // Go through each piece and make sure they exist
            for (const gamePiece of gamePieces) {
                // Remove styles from the game piece
                removeClassApplied(gamePiece);
                
                let gamepieceText = exposed.getPieceText(gamePiece);
                if (!gamepieceText) {
                    continue;					
                }

                // Remove styles from the game piece text
                removeClassApplied(gamepieceText);
                
                gamepieceText.innerHTML = '';
            }
        }
        
        // Reset the list that keeps track of the css classes that are applied
        settings.stylesApplied = [];
    }
    
    /**
    * FUNCTION: getPieceId
    * USE: Returns the Id for the given game piece
    * @param element: JavaScript element representing a game piece
    * @return: Integer game piece id.
    */		
    exposed.getPieceId = function(element) {
        let gamePieceId = element.getAttribute(settings.dataNameGamePieceId);
        if (gamePieceId) {
            gamePieceId = Number(gamePieceId);
        }
        return gamePieceId;
    }
    
    /**
    * FUNCTION: disable
    * USE: Marks a game piece as disabled
    * @param element: JavaScript element representing a game piece
    * @param makeDisabled: Determines if the element should be 
    *    marked as disabled or not
    * @return: N/A.
    */		
    exposed.disable = function(element, makeDisabled = true) {
        if (makeDisabled) {
            exposed.applyClass(element, settings.classNameDisabled);
        } else {
            removeClassApplied(element, settings.classNameDisabled);
        }
    }

    /**
    * FUNCTION: isDisabled
    * USE: Determines if an element is marked as disabled
    * @param element: JavaScript element representing a game piece
    * @return: True if an element is marked as disabled, false otherwise.
    */		
    exposed.isDisabled = function(element) {
        return hasClass(element, settings.classNameDisabled);
    }	

    /**
    * FUNCTION: highlight
    * USE: Highlights a gameboard element
    * @param element: JavaScript element representing a game piece
    * @return: N/A.
    */
    exposed.highlight = function(element) {
        exposed.applyClass(element, settings.classNameHighlight);
    }	
    
    /**
    * FUNCTION: applyClass
    * USE: Adds a css class to an element
    * @param element: JavaScript element
    * @param cssClass: String css class name
    * @return: N/A.
    */		
    exposed.applyClass = function(element, cssClass) {
        let added = addClass(element, cssClass);

        // Keep track of the classes added because 
        // they get automatically removed in the 'clear' function
        if (added) {
            settings.stylesApplied.push(cssClass);
        }
    }
    
// -- Private data --
    let addClass = function(element, cssClass) {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && !hasClass(element, cssClass)) {
            element.classList.add(cssClass)
            modified = true;
        }
        return modified;
    }
    
    let removeClass = function(element, cssClass) {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && hasClass(element, cssClass)) {
            element.classList.remove(cssClass);
            modified = true;
        }
        return modified;
    }

    let hasClass = function(element, cssClass) {
        cssClass = settings.cleanClassName(cssClass);
        return element.classList.contains(cssClass);
    }

    let removeClassApplied = function(element, stylesApplied = null) {
        let customRemoval = false;

        // Check to see if the element has any of 
        // the styles applied to it and remove them
        if (!stylesApplied) {
            stylesApplied = settings.stylesApplied
        } else { 
            if (!Array.isArray(stylesApplied)) {
                stylesApplied = [stylesApplied];
            }
            customRemoval = true;
        }

        for (let x = stylesApplied.length-1; x >= 0; --x) {
            let style = stylesApplied[x];
            let removed = removeClass(element, style);

            if (removed) {
                stylesApplied.splice(x, 1);

                // If this is a custom removal, also remove 
                // the item from the global list
                if (customRemoval) {
                    let index = settings.stylesApplied.indexOf(style);
                    if (index > -1) {
                        //settings.stylesApplied.splice(x, 1);
                    }
                }
            }
        }
    }
}(Game.board)); // http://programmingnotes.org/