// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  elements.js
//    Description: Javascript that handles game elements
// ============================================================================ 

/**
* NAMESPACE: Game.element
* USE: Handles game elements.
*/

// Declare the namespace
registerNamespace('Game.element');

// Declare the namespace members
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
    
    // Set class names and other shared data
    const settings = {	
    };
    exposed.settings = settings;
    
    /**
    * FUNCTION: get
    * USE: Returns the descendant of the object on
    * 	which this method was called.
    * @param selector: The selector to find the element
    * @param element: JavaScript element on which to search
    * @return: The descendant found in the given element.
    */
    exposed.get = (selector, element = document) => {
        return element.querySelector(selector);
    }
    
    /**
    * FUNCTION: getAll
    * USE: Returns all the descendant of the object on
    * 	which this method was called.
    * @param selector: The selector to find the element
    * @param element: JavaScript element on which to search
    * @return: All descendants found in the given element.
    */	
    exposed.getAll = (selector, element = document) => {
        return element.querySelectorAll(selector);
    }
    
    /**
    * FUNCTION: getGameActionsPanel
    * USE: Returns the game 'current actions' panel (div).
    * @return: The action panel.
    */	
    exposed.getGameActionsPanel = () => {
        return exposed.get('#pnlGameActions');
    }
    
    /**
    * FUNCTION: getSkipMoveButton
    * USE: Returns the 'skip move' button.
    * @return: The skip move button.
    */		
    exposed.getSkipMoveButton = () => {
        return exposed.get('#btnSkipMove');
    }
    
    /**
    * FUNCTION: getStartNewGameButton
    * USE: Returns the 'new game' button.
    * @return: The new game button.
    */	
    exposed.getStartNewGameButton = () => {
        return exposed.get('#btnStartNewGame');
    }
    
    /**
    * FUNCTION: getLoadGameButton
    * USE: Returns the 'load game' button.
    * @return: The load game button.
    */	
    exposed.getLoadGameButton = () => {
        return exposed.get('#btnLoadGame');
    }
    
    /**
    * FUNCTION: getSaveGameButton
    * USE: Returns the 'save game' button.
    * @return: The save game button.
    */	
    exposed.getSaveGameButton = () => {
        return exposed.get('#btnSaveGame');
    }

    /**
    * FUNCTION: getAddNewPlayerButton
    * USE: Returns the 'new player' button.
    * @return: The new player button.
    */	
    exposed.getAddNewPlayerButton = () => {
        return exposed.get('#btnAddNewPlayer');
    } 
    
    exposed.populateDropdown = (dropdown, items) => {
        // Clear existing entries
        for (let index = dropdown.options.length-1; index >= 0; --index) {
            dropdown.remove(dropdown.options[index]);
        }
        // Add new entries
        for (const item of items) {
            let option = document.createElement('option');
            for (let prop in item) {
                option[prop] = item[prop];
            }
            dropdown.appendChild(option);         
        }        
    }

    exposed.setSelectedValue = (dropdown, value) => {
        dropdown.selectedIndex = -1;
        value = value != null ? String(value) : value;
        for (let index = 0; index < dropdown.options.length; ++index) {
            if (dropdown.options[index].value === value) {
                dropdown.selectedIndex = index;
                break;
            }
        }
    }     

    exposed.tryFocus = (element) => {
        if (!element.focus) {
            let e = element;
            while (e.frameElement !== null) {e = e.parent;}
            if (e.parent) {
                element = e.parent;
            }
        }
        if (element.setAttribute) {
            element.setAttribute('tabindex', 0);
        }
        if (element.focus) {
            element.focus({preventScroll:true});
        }
    }

    exposed.setSelectionRange = (element, selectionStart, selectionEnd) => {
        let hasFocus = document.activeElement === element;
        if (element.createTextRange) {
            let range = element.createTextRange();
            range.collapse(true);
            range.moveEnd('character', selectionEnd);
            range.moveStart('character', selectionStart);
            range.select();
        } else {
            if (element.selectionStart) {
                if (!hasFocus) {
                    element.focus();
                }
                element.setSelectionRange(selectionStart, selectionEnd); 
            } else {
                if (!hasFocus) {
                    element.focus();
                }
            } 
        }
    }    
    
    exposed.setCaretPosition = (element, caretPos) => {
        exposed.setSelectionRange(element, caretPos, caretPos);
    } 
    
    exposed.getActiveElement = () => {
        let focused = document.activeElement;
        if (!focused || focused == document.body)
            focused = null;
        else if (document.querySelector)
            focused = document.querySelector(':focus');
        return focused;        
    }

    exposed.resetCursor = (element) => {
        try {
            element.value = element.value;
            element.scrollLeft = 0;
            if (/Edge/.test(navigator.userAgent)) {
                element.selectionStart = 0;
            }
        } catch (e) { }        
    }

    exposed.resetScroll = (element) => {
        element.scrollTop = 0;        
    }    

    exposed.scrollToElement = (target, behavior = 'smooth') => {
        if (typeof target === 'string') {
            target = document.querySelector(target);
        }
        target.scrollIntoView({
            behavior: behavior,
            block: 'start'
        });
    }    

    exposed.setVisiblility = (element, makeVisible = true) => {
        if (!element) {
            return;
        }	
        if (makeVisible) {
            element.style.visibility = 'visible';
        } else {
            element.style.setProperty('visibility', 'hidden', 'important');
        }
    }
    
    exposed.setDisplayVisiblility = (element, makeVisible = true) => {
        if (!element) {
            return;
        }
        if (makeVisible) {
            element.style.display = null;
            if (!isVisible(element)) {
                element.style.display = 'block';
            }
        } else {
            element.style.setProperty('display', 'none', 'important');
        }
    }

    exposed.addClass = (element, cssClass) => {
        cssClass = exposed.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && !exposed.hasClass(element, cssClass)) {
            element.classList.add(cssClass)
            modified = true;
        }
        return modified;
    }

    exposed.removeClass = (element, cssClass) => {
        cssClass = exposed.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && exposed.hasClass(element, cssClass)) {
            element.classList.remove(cssClass);
            modified = true;
        }
        return modified;
    }

    exposed.hasClass = (element, cssClass) => {
        cssClass = exposed.cleanClassName(cssClass);
        return element.classList.contains(cssClass);
    }
    
    exposed.cleanClassName = (str) => {
        return str ? str.trim().replace('.', '') : '';
    }
    
// -- Private data --
    let isVisible = (elem) => {
        if (!(elem instanceof Element)) throw Error('DomUtil: elem is not an element.');
        const style = getComputedStyle(elem);
        if (style.display === 'none') return false;
        if (style.visibility !== 'visible') return false;
        if (style.opacity < 0.1) return false;
        if (elem.offsetWidth + elem.offsetHeight + elem.getBoundingClientRect().height +
            elem.getBoundingClientRect().width === 0) {
            return false;
        }
        const elemCenter   = {
            x: elem.getBoundingClientRect().left + elem.offsetWidth / 2,
            y: elem.getBoundingClientRect().top + elem.offsetHeight / 2
        };
        if (elemCenter.x < 0) return false;
        if (elemCenter.x > (document.documentElement.clientWidth || window.innerWidth)) return false;
        if (elemCenter.y < 0) return false;
        if (elemCenter.y > (document.documentElement.clientHeight || window.innerHeight)) return false;
        let pointContainer = document.elementFromPoint(elemCenter.x, elemCenter.y);
        do {
            if (pointContainer === elem) return true;
        } while (pointContainer = pointContainer.parentNode);
        return false;
    }    
}(Game.element)); // http://programmingnotes.org/