// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  events.js
//    Description: Javascript that registers the game events
// ============================================================================ 

/**
* NAMESPACE: Game.events
* USE: Handles registering the game events.
*/

// Declare the namespace
registerNamespace('Game.events');

// Declare the namespace members
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
    
    // Set class names and other shared data
    const settings = {	
    };
    exposed.settings = settings;
    
    exposed.init = () => {
        // Register button clicks for the gameboard pieces
        exposed.registerGamePieceClicks();
        
        // New game button clicks
        exposed.registerNewGameButtonClicks();

        // Skip move button clicks
        exposed.registerSkipMoveButtonClicks();

        // Save game button
        exposed.registerSaveGameButtonClicks();

        // Load game button
        exposed.registerLoadGameButtonClicks();

        exposed.registerPlayerModalButtonClicks();
    }

    /**
    * FUNCTION: registerPlayerModalButtonClicks
    * USE: Registers button clicks for the player modal
    * @return: N/A.
    */     
    exposed.registerPlayerModalButtonClicks = () => {   
        exposed.register(Game.element.getAddNewPlayerButton(), ['click'], (eventClick) => {
            if (Game.status.hasStarted() && !Game.status.hasCompleted()) {
                Game.functions.displayNotice('Finish the match before adding a new player!', Notice.type.error);
                return;                
            }
            Game.players.modal.show(Game.players.modal.view.add);           
        });
    }


    /**
    * FUNCTION: registerNewGameButtonClicks
    * USE: Registers button clicks for the new game button
    * @return: N/A.
    */    
    exposed.registerNewGameButtonClicks = () => {
        exposed.register(Game.element.getStartNewGameButton(), ['click'], (eventClick) => {
            try {
                if (!Game.status.hasStarted() || !Game.players.getStarting() 
                    || !Game.players.isValidPlayer(Game.players.get(Game.players.getStarting()))) {
                    Game.functions.displayNotice('Please select a starting player!', Notice.type.error);
                    return;
                }
        
                // Check to see if any moves have been made before starting a new match
                let match = Game.status.getCurrentMatch();
                if (match.getTotalMovesRemaining() < match.getTotalMovesPossible()) {
                    Game.functions.startNewMatch();
                    Game.functions.displayNotice('Starting new match...', Notice.type.success, 1000);
    
                    // Make a move if the current player is an AI player
                    Game.functions.makeMove();                
                }
            } catch (e) {
                Game.functions.displayNotice(e.message, Notice.type.error);
            }
        });       
    }  
    
    /**
    * FUNCTION: registerSkipMoveButtonClicks
    * USE: Registers button clicks for the skip move button
    * @return: N/A.
    */      
    exposed.registerSkipMoveButtonClicks = () => {
        exposed.register(Game.element.getSkipMoveButton(), ['click'], (eventClick) => {
            let match = Game.status.getCurrentMatch();
    
            if (!Game.status.hasStarted(match)) {
                return;
            } else if(Game.status.hasCompleted(match)) {
                Game.functions.displayNotice('Start a new game to play again!', Notice.type.warning);
                return;
            }
    
            match.moveNextPlayer();
            Game.functions.refreshPlayerMoves(match);

            // Make a move if the current player is an AI player
            Game.functions.makeMove();            
        });      
    }  
    
    /**
    * FUNCTION: registerSaveGameButtonClicks
    * USE: Registers button clicks for the save game button
    * @return: N/A.
    */     
    exposed.registerSaveGameButtonClicks = () => {
        exposed.register(Game.element.getSaveGameButton(), ['click'], (eventClick) => {
            if (!Game.status.hasStarted()) {
                Game.functions.displayNotice('Please start a match before saving!', Notice.type.warning);
                return;
            }

            // Save the game
            Game.functions.saveGame();		
        });
    } 
    
    /**
    * FUNCTION: registerLoadGameButtonClicks
    * USE: Registers button clicks for the load game button
    * @return: N/A.
    */     
    exposed.registerLoadGameButtonClicks = () => {
        exposed.register(Game.element.getLoadGameButton(), ['click'], async (eventClick) => { 
            await Game.functions.loadGame();				
        });	     
    }     

    /**
    * FUNCTION: registerGamePieceClicks
    * USE: Registers button clicks for the game piece button
    * @return: N/A.
    */     
    exposed.registerGamePieceClicks = () => {
        Game.board.getPieces(document).forEach((gamePiece, index) => {
            exposed.register(gamePiece, ['click'], (eventClick) => {
                let match = Game.status.getCurrentMatch();
    
                if (Game.board.isDisabled(gamePiece) 
                    || !Game.players.getStarting() || Game.status.hasCompleted(match)) {
                    if (!Game.players.getStarting()) {
                        Game.functions.displayNotice('Please select a starting player!', Notice.type.error);
                    } else if (Game.status.hasCompleted(match)) {
                        Game.functions.displayNotice('Start a new game to play again!', Notice.type.warning);
                    } else {
                        //alert('Disabled!');
                    }
                    return;
                }
                
                Game.functions.markSelection(gamePiece, match);

                // Make a move if the current player is an AI player
                Game.functions.makeMove();
            });	
        });        
    }

    /**
    * FUNCTION: registerGamePieceClicks
    * USE: Registers button clicks for the game piece button
    * @return: N/A.
    */       
    exposed.registerRadioButtonClicks = () => {
        // Make sure the radio group is setup with the selected option slider div and other options
        // And add button clicks to animate the radio button slider	
        Slider.init();

        Slider.getRadioButtons(document).forEach((radioButton, index) => {
            exposed.register(radioButton, ['click'], (eventClick) => {
                try {
                    let match = Game.status.getCurrentMatch();
    
                    if (Game.status.hasStarted(match) && !Game.status.hasCompleted(match)) {
                        Game.functions.displayNotice('Please finish the match before setting a new starting player!', Notice.type.warning);
                        return;
                    }
        
                    Game.functions.startNewMatch(radioButton.value);
    
                    // Make a move if the current player is an AI player
                    Game.functions.makeMove();
                } catch (e) {
                    Game.functions.displayNotice(e.message, Notice.type.error);
                }
            });
        });

        let editContainerSelector = '.player-section-edit';
        let editSections = Game.element.getAll(editContainerSelector);
        editSections.forEach((section, index) => {
            let playerId = section.getAttribute('data-playerId');
            let player = Game.players.get(playerId);
            
            // Get the edit player button
            let buttonEdit = Game.element.get('.playerButtonEdit', section);
            exposed.register(buttonEdit, ['click'], (eventClick) => {
                eventClick.preventDefault();
                if (Game.status.hasStarted() && !Game.status.hasCompleted()) {
                    Game.functions.displayNotice('Finish the match before editing player!', Notice.type.warning);
                    return;                
                }

                Game.players.modal.show(Game.players.modal.view.update, player);
            });

            // Get the remove player button
            let buttonRemove = Game.element.get('.playerButtonRemove', section);
            exposed.register(buttonRemove, ['click'], (eventClick) => {
                eventClick.preventDefault();
                if (Game.status.hasStarted() && !Game.status.hasCompleted()) {
                    Game.functions.displayNotice('Finish the match before removing player!', Notice.type.warning);
                    return;                
                }                
                try {
                    Modal.showDialog({
                        title: `Remove ${player.name}?`,
                        body: `Are you sure you want to remove ${player.name} from the game?`,
                        buttons: {
                            cancel: {text: 'Cancel'}, 
                            ok: {text: `❌ Remove ${player.name}`, onClick: (event) => removePlayer()},            
                        },
                    }); 
                    
                    let removePlayer = () => {
                        if (!Game.players.remove(player)) {
                            throw new Error(`There was a problem removing ${player.name} from the game!`);
                        }
        
                        Game.functions.displayNotice(`${player.name} successfully removed from the game`, Notice.type.success);
                        Game.functions.createPlayerRadioButtons(Game.players.get(), Game.players.getStarting());
                    }
                } catch (e) {
                    Game.functions.displayNotice(e.message, Notice.type.error);
                }
            });   
            
            exposed.register(section, ['click'], (eventClick) => {
                eventClick.preventDefault();                
            }); 
        });                             
    }

    exposed.register = (element, eventNames, func) => {
        eventNames.forEach((eventName, index) => {
            element.removeEventListener(eventName, func);
            element.addEventListener(eventName, func);
        });
    }    
    
// -- Private data --
    
}(Game.events)); // http://programmingnotes.org/