// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  functions.js
//    Description: Javascript that handles general gameplay related functions 
//       for the game
// ============================================================================ 

/**
* NAMESPACE: Game.functions
* USE: Handles general gameplay related functions
*/

// Declare the namespace
registerNamespace('Game.functions');

// Declare the namespace members
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;

    exposed.isNull = (item) => {
        return undefined === item || null === item;
    }

    exposed.isEmpty = (str) => {
        return exposed.isNull(str) || String(str).trim().length < 1;
    }    

    exposed.makeMove = (player = null, match = null, delay = 200) => {
        if (!match) {
            match = Game.status.getCurrentMatch();
        }
        if (Game.status.hasCompleted(match)) {
            return;
        }
        let playerSpecified = true;
        if (!player) {
            playerSpecified = false;
            player = Game.players.get(match.getCurrentPlayer());
        }
        if (player && (playerSpecified || player.isAiPlayer)) {
            setTimeout(() => {
                let move = Game.ai.getMove(player.getId(), player.difficulty);
                if (move != null) {
                    let gamePiece = Array.from(Game.board.getPieces())
                        .find((gamePiece) => Game.board.getPieceId(gamePiece) == move);
                    if (gamePiece) {
                        elementClick(gamePiece);
                    }
                }
            }, delay);
        }
    }
    
    /**
    * FUNCTION: resetGameArea
    * USE: Resets the game interface to its default state.
    * @return: N/A.
    */	
    exposed.resetGameArea = () => {
        Game.board.clear();
        Game.element.setVisiblility(Game.element.getGameActionsPanel(), true);
        Game.element.setDisplayVisiblility(Game.element.getSkipMoveButton(), true);
        exposed.showPlayerSection(PlayerSection.playerMoves);
    }
    
    /**
    * FUNCTION: disableBoard
    * USE: Marks all the gampieces as 'disabled'.
    * @param makeDisabled: True if the board should be disabled, false otherwise
    * @return: N/A.
    */	
    exposed.disableBoard = (makeDisabled = true) => {
        Game.board.getPieces(document).forEach((element, index) => {
            Game.board.disable(element, makeDisabled);
        });
    }
    
    /**
    * FUNCTION: displayNotice
    * USE: Displays a notice to the screen with the specified text.
    * @param text: The text to be displayed
    * @param type: The notice type
    * @param duration: The notice duration
    * @return: N/A.
    */	
    exposed.displayNotice = (text, type = Notice.type.success, duration = 5000) => {
        let element = Game.element.get(Notice.settings.classNameNotice);
        Notice.display(element, text, type, duration);
    }
    
    /**
    * FUNCTION: disableRadioButtons
    * USE: Marks all the radio buttons as 'disabled'.
    * @param makeDisabled: True if the buttons should be disabled, false otherwise
    * @return: N/A.
    */		
    // let radioButtonsDisabled = null;
    exposed.disableRadioButtons = (makeDisabled = true) => {
        // if (radioButtonsDisabled !== null && radioButtonsDisabled == makeDisabled) {
        // 	return;
        // }
    
        Slider.getRadioButtons(document).forEach((radioButton, index) => {
            let radioGroup = Slider.getGroupForButton(radioButton);
            let radioLabel = Slider.getLabelForButton(radioButton, radioGroup)
            let selectedItem = Slider.getSelectedItem(radioGroup);
    
            Game.board.disable(radioLabel, makeDisabled);
            Game.board.disable(selectedItem, makeDisabled);
            Game.board.disable(radioButton, makeDisabled);
    
            if (makeDisabled) {
                radioButton.setAttribute('disabled', makeDisabled);
            } else {
                radioButton.removeAttribute("disabled");
            }
        });
        // radioButtonsDisabled = makeDisabled;
    }
    
    /**
    * FUNCTION: createPlayerRadioButtons
    * USE: Creates radio buttons from the given players.
    * @param players: Players of the game
    * @return: N/A.
    */	
    exposed.createPlayerRadioButtons = (players, starting = null) => {
        // radioButtonsDisabled = null;
    
        // Find the players radio button container
        let buttonContainer = Game.element.get('.section-radio-button');
        if (!buttonContainer) {
            return;
        }
    
        // Clear existing elements in this container
        let previousRadioGroups = Slider.getRadioGroups(buttonContainer);
        for (const group of previousRadioGroups) {
            buttonContainer.removeChild(group);
        }
    
        // Exit if no players are specified
        if (!players) {
            return;
        } else if (!Array.isArray(players)) {
            players = [players];
        }
    
        if (players.length < 1) {
            return;
        }
    
        // Create the radio group
        let radioGroup = document.createElement('div');
        Game.element.addClass(radioGroup, Slider.settings.classNameRadioGroup); 
        Game.element.addClass(radioGroup, Slider.settings.classNameRadioColumn);
        
        let radioIdName = 'rdoStartingPlayer';
    
        // Create a radio button for each player
        for (const player of players) {
            let checked = starting != null ? starting === player.getId() : false;
            
            // Create the radio button
            let radioButton = document.createElement('input');
            radioButton.type = 'radio';
            Game.element.addClass(radioButton, Slider.settings.classNameRadioButton);
            radioButton.id = `${radioIdName}_${player.getId()}`;
            radioButton.value = String(player.getId());
            radioButton.name = `${radioIdName}_Group`;
            radioButton.checked = checked;
    
            // Create the radio option label
            let radioButtonLabel = document.createElement('label');
            Game.element.addClass(radioButtonLabel, Slider.settings.classNameRadioButtonLabel);
            radioButtonLabel.htmlFor = radioButton.id;
            radioButtonLabel.innerHTML = getRadioButtonHtml(player);
    
            // Add the radio button and option label to the radio group
            radioGroup.appendChild(radioButton);
            radioGroup.appendChild(radioButtonLabel);
        }

        buttonContainer.appendChild(radioGroup);
    
        // Register game button clicks for the radio buttons
        Game.events.registerRadioButtonClicks();
    }

    let getRadioButtonHtml = (player) => {
        let html = `${player.getRadioButtonHtml()}`;
        let data = `data-playerId=${player.getId()}`;
        html += `
            <section class="player-section-edit" ${data}>
                <div title="Remove: ${player.name}"
                    class="inline button image icon remove playerButtonRemove">                    
                </div>
                <div title="Edit: ${player.name}"
                    class="inline button image icon edit playerButtonEdit">                     
                </div>                
            </section>
        `;
        return html;
    }
    
    let totalMovesPossible = null;
    exposed.getTotalMovesPossible = () => {
        if (!totalMovesPossible) {
            totalMovesPossible = Game.board.getPieces(document).length;
        }
        return totalMovesPossible;
    }
    
    exposed.refreshPlayerMoves = (match = null) => {
        if (!match) {
            match = Game.status.getCurrentMatch();
        }
    
        if (Game.status.hasCompleted(match)) {
            return;
        }
    
        // Get the current player
        exposed.showPlayerIconSection(PlayerIconSection.currentMove, match.getCurrentPlayer());
    
        // Get the next player
        exposed.showPlayerIconSection(PlayerIconSection.nextMove, match.peekNextPlayer());
    }
    
    exposed.showPlayerIconSection = (type, player) => {
        let element = exposed.getPlayerIconSection(type);
        if (!element) {
            return;
        }
    
        if (player) {
            // Get the player 
            let currentPlayer = Game.players.get(player);

            // show their icon
            element.innerHTML = `${currentPlayer.getGamePieceHtml()}`;	
        } else {
            // show their icon
            element.innerHTML = '';	
        }
    }
    
    const PlayerSection = Object.freeze({none:0, playerMoves: 1, winner:2, tie: 3});
    exposed.PlayerSection = PlayerSection;
    exposed.showPlayerSection = (type = PlayerSection.none) => {
        
        // Go through each section and only show the one specified
        for (const prop in PlayerSection) {
            // Get the section type
            const section = PlayerSection[prop];
            let setVisible = (section == type);
    
            // Get the section and make sure it exists
            let element = exposed.getPlayerSection(section);
            if (!element) {
                continue;
            }
    
            // Change its visibility
            Game.element.setDisplayVisiblility(element, setVisible)
        }
    }
    
    exposed.getPlayerSection = (type = PlayerSection.none) => {
        let sectionSelector = null;
        let element = null;
    
        switch (type) {
            case PlayerSection.playerMoves:
                sectionSelector = '.section-player-moves';
                break;					
            case PlayerSection.winner:
                sectionSelector = '.section-player-winner';
                break;
            case PlayerSection.tie:
                sectionSelector = '.section-player-tie';			
                break;
        }
    
        if (sectionSelector) {
            element = Game.element.get(sectionSelector);
        }
    
        return element;
    }
    
    const PlayerIconSection = Object.freeze({none:0, currentMove:1, nextMove:2, winner:3});
    exposed.PlayerIconSection = PlayerIconSection;
    exposed.getPlayerIconSection = (type = PlayerIconSection.none) => {
        let sectionSelector = null;
        let element = null;
    
        switch (type) {
            case PlayerIconSection.currentMove:
                sectionSelector = '.player-current-move';
                break;
            case PlayerIconSection.nextMove:
                sectionSelector = '.player-next-move';
                break;			
            case PlayerIconSection.winner:
                sectionSelector = '.player-winner';
                break;
        }
    
        if (sectionSelector) {
            element = Game.element.get(sectionSelector);
        }
    
        return element;
    }
    
    exposed.generateId = () => {
        return String((Math.random() * 100)).replace('.', '');
    }
    
    exposed.highlightGamePieces = (pieceIds) => {
        if (!pieceIds || pieceIds.length < 1) {
            return;
        }
        
        Game.board.getPieces(document).forEach((gamePiece, index) => {
            let pieceId = Game.board.getPieceId(gamePiece);
            if (pieceId && pieceIds.indexOf(pieceId) > -1) {
                Game.board.highlight(gamePiece);
            }
        });  
    }
    
    exposed.markSelection = (gamePiece, match) => {
        // Get the game piece id
        let pieceId = Game.board.getPieceId(gamePiece);
        if (!pieceId) {
            throw new Error('Game piece id is not specified');
        }
    
        // Get the player that made the move
        let player = match.getCurrentPlayer();
    
        // Mark the piece with the player name/icon
        exposed.markPieceAsSelected(gamePiece, player);
        
        // Add this move for the current player
        match.addMove(pieceId, player);
    
        // Check to see if a winner is found
        exposed.checkWin(match);
    }
    
    exposed.markPieceAsSelected = (gamePiece, player) => {
        let pieceText = Game.board.getPieceText(gamePiece);

        // Disable the piece from being selected again
        Game.board.disable(gamePiece);	

        if (!Game.players.isValidPlayer(player)) {
            player = Game.players.get(player);
        }
        
        // Set the html for the game piece
        pieceText.innerHTML = `${player.getGamePieceHtml()}`;	
    }
    
    exposed.startNewMatch = (startingPlayer = null) => {
        if (startingPlayer) {
            // Set the starting player
            Game.players.setStarting(startingPlayer);
        }
    
        // Get the players for the game
        let players = Game.players.get();
    
        // Get the total # of moves possible
        let totalMovesPossible = exposed.getTotalMovesPossible();
    
        // Get the starting player
        startingPlayer = Game.players.getStarting();
    
        // Start a new match
        Game.status.startNewMatch(players, totalMovesPossible, startingPlayer);
    
        exposed.resetGameArea();
        exposed.disableRadioButtons(true);
    
        // Refresh current move
        exposed.refreshPlayerMoves();

        exposed.createMatchHistoryPanel(Game.status.getMatches());
    }
    
    exposed.checkWin = (match = null) => {
        if (!match) {
            match = Game.status.getCurrentMatch();
        }
        
        try {
            let matchStatus = match.checkWin();
    
            switch (matchStatus.result) {
                case Game.status.result.winnerFound:
                    //console.log(match.getWinner());
                    exposed.highlightWinner(match);
                    break;
                case Game.status.result.tieGame:
                    exposed.displayNotice('Players Tied', Notice.type.warning);
                    exposed.showPlayerSection(PlayerSection.tie);
                    break;
                case Game.status.result.none:
                    //console.log('More moves are available');
                    // Refresh current move
                    exposed.refreshPlayerMoves(match);
                    break;
                default:
                    throw new Error('Unknown match result status');
                    break;
            }
    
            if (Game.status.hasCompleted(match)) {
                exposed.disableRadioButtons(false);
                Game.element.setDisplayVisiblility(Game.element.getSkipMoveButton(), false);
                exposed.createMatchHistoryPanel(Game.status.getMatches());
            }		
    
            //console.log(matchStatus);
        } catch (e) {
            //console.error(e.name, e.message);
            exposed.displayNotice(e.message, Notice.type.error, null);
        }	
    }
    
    exposed.highlightWinner = (match) => {
        if (!Game.status.hasCompleted(match)) {
            throw new Error('Game has not ended');
        } else if (!match.getWinner()) {
            throw new Error('Game has no winner');
        }
    
        exposed.disableBoard(true);	
    
        // Get the winning player
        let player = Game.players.get(match.getWinner().playerId);
    
        // Get the winning combo so we can highight those gamepieces
        let winningCombo = match.getWinner().winningCombo;
    
        //console.log(match);
    
        // Highlight the winning gamepieces
        exposed.highlightGamePieces(winningCombo);
    
        exposed.showPlayerSection(PlayerSection.winner);
        exposed.showPlayerIconSection(PlayerIconSection.winner, player);

        if (player) {
            exposed.displayNotice(`${player.name} won the game!`, Notice.type.success);
        }
    }
    
    exposed.createMatchHistoryPanel = (matches) => {
        // Find the collapsible container
        let collapsibleContainer = Game.element.get('.collapsible-container');
        if (!collapsibleContainer) {
            return;
        }
    
        // Get the instructions and determine if it should be visible or invisible
        let collapsibleInstructions = Game.element.get('.collapsible-empty-text', collapsibleContainer);
        Game.element.setDisplayVisiblility(collapsibleInstructions, true);
    
        // Clear existing panels in this container
        let previousPanels = Collapsible.getPanels(collapsibleContainer);
        previousPanels.forEach((panel, index) => {
            for (const prop in panel) {
                let element = panel[prop];
                if (!element) {
                    continue;
                }
                collapsibleContainer.removeChild(element);
            }
        });
    
        // Exit if no matches are specified
        if (!matches) {
            return;
        } else if (!Array.isArray(matches)) {
            matches = [matches];
        }
    
        // Dont show the current match unless its been completed
        let currentMatch = Game.status.getCurrentMatch();
        if (currentMatch) {
            if (!Game.status.hasCompleted(currentMatch)) {
                matches = matches.filter((x) => x.getId() != currentMatch.getId());
            }		
        }
        
        // Exit if there are no matches
        if (matches.length < 1) {
            return;
        }
    
        // Sort the matches in descending order
        matches = Sort.by( matches,
            { value: (x) => x.getTimeStart().getTime(), desc: true }
        );
    
        matches.forEach((match, index) => {
            // collapsible header
            let header = document.createElement('div');
            Game.element.addClass(header, Collapsible.settings.classNameCollapsible);
            Game.element.addClass(header, Collapsible.settings.classNameHeader);
            
            // collapsible header text
            let headerText = document.createElement('div');
            Game.element.addClass(headerText, Collapsible.settings.classNameHeaderText);
    
            // Show the winner name or if it was tie game
            headerText.innerHTML = exposed.getHistoryTitle(match);
            header.title = headerText.innerText;
    
            // collapsible button
            let headerButton = document.createElement('div');
            Game.element.addClass(headerButton, Collapsible.settings.classNameButton);
            
            header.appendChild(headerText);
            header.appendChild(headerButton);
    
            // collapsible body
            let body = document.createElement('div');
            Game.element.addClass(body, Collapsible.settings.classNameBody);
    
            body.innerHTML = ` 
                <div style="padding: 10px 0 10px 0;">
                    ${exposed.getHistoryBody(match)}
                </div>
            `;
    
            collapsibleContainer.appendChild(header);
            collapsibleContainer.appendChild(body);
        });
    
        // Make sure the collapsible is setup
        Collapsible.init(collapsibleContainer);
    
        Game.element.setDisplayVisiblility(collapsibleInstructions, false);

        Game.element.resetScroll(Game.element.get('.section-game-history-container'));
    }
    
    exposed.getHistoryBody = (match) => {
        let duration = match.getDuration();
    
        let details = `
            <div>
                Date: ${match.getTimeStartString('d')}
            </div>	
            <div>
                Time: ${match.getTimeStartString('t')} - ${match.getTimeEndString('t')}
            </div>		
            <div>
                Duration: ${duration.totalSeconds > 0 ? duration.toString() : 'N/A'}
            </div>		
        `
    
        // Get info for the players that were in the game
        let players = exposed.getPlayersInMatch(match);
    
        // Sort the players by the winner, # moves, and then by the player name
        players = Sort.by(players,
            [
                { value: (x) => x.isWinner, desc: true },
                { value: (x) => x.moves.length, desc: true },
                { value: (x) => x.player ? x.player.name : '' },
            ]);
    
        details = exposed.appendText(details, `
            <div class="matches table-container">
                <table class="matches table">
                    <tr class="matches table-row">
                        <th class="matches table-header">Player</th>
                        <th class="matches table-header"># Moves</th>
                        <th class="matches table-header">Winner</th>
                    </tr>
        `);
    
        players.forEach((playerInfo, index) => {
            let playerId = playerInfo.playerId;
            let player = playerInfo.player;
            let playerMoves = playerInfo.moves;
            let isWinner = playerInfo.isWinner;
            let info = playerInfo.info;

            let nameDescription = ``;
            if (info && info.isAiPlayer) {
                nameDescription = `
                    <span class="matches table-data name-description">
                        CPU: ${Game.ai.getDescription(info.difficulty)}
                    </span> 
                `;
            }
    
            details = exposed.appendText( details, `
                    <tr class="matches table-row">
                        <td class="matches table-data">
                            ${exposed.getPlayerName(player)}
                            ${nameDescription}                      
                        </td>
                        <td class="matches table-data">
                            ${playerMoves.length}
                        </td>
                        <td class="matches table-data">
                            <div class="matches winner">
                                ${isWinner ? '✔️' : '❌'}
                            </div>
                        </td>
                    </tr>
            `);
        });
    
        details = exposed.appendText(details, `
                </table>
            </div>
        `);
    
        return details
    }
    
    // Show the winner name or if it was tie game
    exposed.getHistoryTitle = (match) => {
        let additional = 'No Winner';
        let matchStatus = match.getStatus();
    
        if (matchStatus) {
            let description = Game.status.getResultDescription(matchStatus.result);
            let winner = match.getWinner();

            if (description.length > 0) {
                additional = description;
            }
    
            if (winner && winner.playerId) {
                additional = exposed.appendText(additional, exposed.getPlayerName(winner.playerId), ': ');
            }		
        }
        
        let title = `Match #${match.getId()}${additional.length > 0 ? ` - ${additional}` : ''}`;
        return title;
    }
    
    exposed.appendText = (str, text, prefixText = '') => {
        str = str || '';
        text = text || '';
        str += `${str.length > 0 ? prefixText : ''}${text.length > 0 ? text : ''}`.trim();
        return str;
    }
    
    exposed.getPlayersInMatch = (match) => {
        let players = [];
        let playerIdWinner = match.getWinner() != null ? match.getWinner().playerId : null;
    
        for (const playerId of match.getPlayers()) {
            let playerInfo = {};
            playerInfo.playerId = playerId;
            playerInfo.moves = match.getMoves(playerId);
            playerInfo.info = match.getPlayerInfo(playerId);
            playerInfo.player = null;
            playerInfo.isWinner = (playerIdWinner == playerId);
    
            let player = Game.players.get(playerId);
            if (Game.players.isValidPlayer(player)) {
                playerInfo.player = player;
            }
    
            players.push(playerInfo);
        }	
        return players;
    }
    
    exposed.getPlayerName = (player, nullPlayerName = '*Removed*') => {
        if (!Game.players.isValidPlayer(player)) {
            player = Game.players.get(player);
        }
        let name = player && player.name ? player.name : nullPlayerName;
        return name;
    }
    
    exposed.reloadGameSave = (gameSaveData) => {
        // Load the game matches
        Game.status.load(gameSaveData, true);	

        // Load the game players
        Game.players.load(gameSaveData);

        // Load the starting player
        try {
            Game.players.setStarting(gameSaveData.startingPlayer);	
        } catch (e) {}        	

        // Get the current match
        let match = Game.status.getCurrentMatch();				
    
        // Reload informtation for the most recent saved game
        exposed.resetGameArea();
        let gamePieces = Array.from(Game.board.getPieces(document));
    
        // Go through each player move and set their piece on the gameboard
        for (let player of match.getPlayers()) {
            for (let move of match.getMoves(player)) {
                let gamePiece = gamePieces.find((gamePiece) => Game.board.getPieceId(gamePiece) == move);
                if (!gamePiece) {
                    continue;
                }
                try {
                    exposed.markPieceAsSelected(gamePiece, player);
                } catch (e) {} 
            }
        }
    
        exposed.LoadRadioAndMatchHistory();
    
        // Set the radio button as the starting player
        let startingPlayer = Game.players.getStarting();
        if (startingPlayer) {
            startingPlayer = String(startingPlayer);
            let radioButtons = Array.from(Slider.getRadioButtons(document));
            let radioButton = radioButtons.find((x) => x.value == startingPlayer);
            if (radioButton) {
                radioButton.checked = true;
                exposed.disableRadioButtons(true);
            }            
        }		
        exposed.checkWin(match);
    }
    
    exposed.saveGame = () => {
        let data = serialize( generateGameSaveData() );
        let filename = generateFilename('gamesave', '.json');
        FileIO.save({
            data: data,
            filename: filename
        });	
    }	

    exposed.loadGame = async () => {
        try {
            // Open a file prompt to get the save game file from the user 
            let files = await FileIO.openDialog({
                contentTypes: '.json',
                allowMultiple: false,
            });	
            
            // Make sure the file is valid
            if (!files || files.length < 1) {
                throw new TypeError('Please select a file!');
            }
    
            // Get the first file in the list
            let file = files[0]; 
            if (file.name.indexOf('.json') < 0) {
                throw new TypeError(`Unable to load. File: '${file.name}' is invalid. Expected a .json file.`);
            }
            
            // Read the file
            let contents = await FileIO.read({
                data: file
            });
    
            // Deserialize and validate the game save data
            let gameSaveData = validateGameSaveData(contents);
    
            exposed.reloadGameSave(gameSaveData);	
    
            exposed.displayNotice('Game successfully loaded!');
        } catch (e) {
            let message = 'Unable to load. Save file is invalid.';
            let duration = 5000;
            if (e instanceof TypeError) {
                message = e.message;
                duration = 10000;
            }
            //console.error(e.name, e.message, e);
            exposed.displayNotice(message, Notice.type.error, duration);			
        }
    }	
    
    exposed.LoadRadioAndMatchHistory = () => {
        // Create radio buttons for the players 
        // to determine who the starting player is
        exposed.createPlayerRadioButtons(Game.players.get());	
    
        // Crate match history panel
        exposed.createMatchHistoryPanel(Game.status.getMatches());	
    }
    
    exposed.arrayMax = (array, selector) => {
        if (typeof selector !== 'function') {
            selector = (item) => item;
        }
        let arrayItem = array.reduce((x, y) => selector.call(this, x) > selector.call(this, y) ? x : y);        
        let value = selector.call(this, arrayItem);
        return value;
    }    	
    
    let getGameSaveDataTypeMaps = () => {
        let typeMaps = {
            players: Game.players.Player,
            matches: Game.status.Match		
        };
        return typeMaps;
    }  
    
    let convertData = (data, typeMaps) => {
        if (typeof data == 'string') {
            data = deserialize(data);
        }
        
        // Go through each property and convert 
        // objects to the specific types 
        for (const prop in typeMaps) {
            let list = data[prop];		
            if (!list) {
                continue;
            }
            let type = typeMaps[prop];
            list.forEach((obj, index) => { 
                data[prop][index] = ctype(obj, type);
            });
        }
        return data
    }

    let serialize = (data) => {
        let serialized = JSON.stringify(data, undefined, 4);
        return serialized;
    }
        
    let deserialize = (data) => {
        let deserialized = JSON.parse(data, (key, value) => parseDate(value));
        return deserialized;
    }

    let parseDate = (value) => {
        if (typeof value === 'string') {
            // Check ISO-formatted string
            let reISO = /^\d{4}-(0[1-9]|1[0-2])-([12]\d|0[1-9]|3[01])([T\s](([01]\d|2[0-3])\:[0-5]\d|24\:00)(\:[0-5]\d([\.,]\d+)?)?([zZ]|([\+-])([01]\d|2[0-3])\:?([0-5]\d)?)?)?$/;
            let a = reISO.exec(value);
            if (a) {
                // if so, Date() can parse it
                return new Date(value);
            }

            // Check Microsoft-format string
            let reMsAjax = /^\/Date\((d|-|.*)\)[\/|\\]$/;
            a = reMsAjax.exec(value);
            if (a) {
                // Parse for relevant portions
                let b = a[1].split(/[-+,.]/);
                return new Date(b[0] ? +b[0] : 0 - +b[1]);
            }
        }
        return value; 		
    } 
    
    let ctype = (obj, type) => {
        let retValue = obj;
        if (!(obj instanceof type)) {
            retValue = Object.create(type.prototype, Object.getOwnPropertyDescriptors(obj));
        }
        return retValue;
    }
    
    let generateFilename = (prefix, extension = 'json') => {
        let now = new Date();
        let time = Game.status.getDateTimeString(now);
        let replaceValue = '-';
        time = time.replace(/\//g, replaceValue).replace(/\s/g, replaceValue).replace(/:/g, replaceValue);
        let filename = `${prefix}_${time}.${extension.replace('.', '')}`
        return filename;
    }
    
    let validateGameSaveData = (contents) => {
        let typeMaps = getGameSaveDataTypeMaps()
        let gameSaveData = convertData(contents, typeMaps)
        
        // Make sure the required properties are in the game save data
        for (let prop in typeMaps) {
            if (!gameSaveData[prop]) {
                throw new TypeError('Unable to load game. Save file data is invalid');
            }		
        }
    
        return gameSaveData;
    }
    
    let generateGameSaveData = () => {
        let data = {
            players: Game.players.get(),
            matches: Game.status.getMatches(),
            startingPlayer: Game.players.getStarting(),
            playersMaxId: Game.players.getMaxId(),
            matchesMaxId: Game.status.getMaxId(),
        };
        return data;
    }	

    let elementClick = (element) => {
        try {
            element.dispatchEvent(new MouseEvent('click'))
        } catch (e) {
            let evt = document.createEvent('MouseEvents')
            evt.initMouseEvent('click', true, true, window, 0, 0, 0, 80,
                                20, false, false, false, false, 0, null)
            element.dispatchEvent(evt)
        }
    } 
 }(Game.functions)); // http://programmingnotes.org/