// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Sept 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  playerModal.js
//    Description: Javascript that handles the the player modal
// ============================================================================ 

/**
* NAMESPACE: Game.players.modal
* USE: Handles the the player modal.
*/

// Declare the namespace
registerNamespace('Game.players.modal');

// Declare the namespace members
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
    
    const ModalView = Object.freeze({add:1, update: 2});
    exposed.view = ModalView;
    exposed.show = (view, player = null) => {
        init();
        clearControls();        

        switch (view) {
            case ModalView.add:
                Modal.getHeaderText(modalElement).innerText = 'Add New Player';
                modalControls.btnEditPlayerSave.innerText = 'Add New';
                break;
            case ModalView.update:
                Modal.getHeaderText(modalElement).innerText = 'Update Player';
                modalControls.btnEditPlayerSave.innerText = 'Update'; 
                if (player) {
                    player = Game.players.get(player);
                }
                if (!Game.players.isValidPlayer(player)) {
                    throw new Error('No player specified')
                }
                break;
            default:
                throw new TypeError(`Unknown view mode: ${view}`);
                break;
        }

        modalControls.hdnModalView.value = view;

        if (player) {
            modalControls.hdnPlayerId.value = player.getId();
            modalControls.txtName.value = player.name;
            modalControls.txtIconUrl.value = player.iconUrl;
            modalControls.chkShowIcon.checked = player.showIcon;
            modalControls.chkIsAiPlayer.checked = player.isAiPlayer;
            if (!Game.functions.isNull(player.difficulty)) {
                Game.element.setSelectedValue(modalControls.cboDifficulty, player.difficulty);
            }
        }
        
        Modal.show(modalElement);
    }

    exposed.hide = () => {
        init();
        Modal.hide(modalElement);
        clearControls(clearDelay);
    }

// -- Private data --
    let showMessage = (text) => {
        modalControls.divErrorMessage.innerHTML = text;
        Game.element.setDisplayVisiblility(modalControls.divErrorMessage, true);
        Game.element.scrollToElement(modalControls.divErrorMessage);
    }

    let hideMessage = () => {
        Game.element.setDisplayVisiblility(modalControls.divErrorMessage, false);
    }    

    let init = () => {
        if (initComplete) {
            return;
        } 
        Modal.init();                
        modalElement = Game.element.get('#mpEditPlayer');
        modalControls = getControls();  
        registerButtonClicks();
        Placeholder.float(modalElement.querySelectorAll('input[type="text"], select'));
        Game.element.populateDropdown(modalControls.cboDifficulty, getDropdownItems(true));
        initComplete = true;
    }

    let getDropdownItems = (addEmpty = false) => {
        let items = [];
        if (addEmpty) {
            items.push({
                text: '',
                value: '',
            }); 
        }       
        let difficulty = Game.ai.difficulty;
        for (const prop in difficulty) {
            let value = difficulty[prop];
            let text = Game.ai.getDescription(value);
            items.push({
                text: text,
                value: value,
            });           
        }        
        return items;
    }   

    let registerButtonClicks = () => {        
        Game.events.register(modalControls.txtIconUrl, ['change', 'keyup'], (event) => { 
            modalControls.chkShowIcon.checked = !Game.functions.isEmpty(modalControls.txtIconUrl.value);
        });        

        Game.events.register(modalControls.txtIconUrl, ['blur'], (event) => { 
            Game.element.resetCursor(modalControls.txtIconUrl);
        });

        Game.events.register(modalControls.txtName, ['blur'], (event) => { 
            Game.element.resetCursor(modalControls.txtName);
        });        

        Game.events.register(modalControls.cboDifficulty, ['change'], (event) => { 
            modalControls.chkIsAiPlayer.checked = !Game.functions.isEmpty(modalControls.cboDifficulty.value);
        });

        Game.events.register(modalControls.btnEditPlayerCancel, ['click'], (eventClick) => { 
            exposed.hide();
        });
        
        Game.events.register(modalControls.btnEditPlayerSave, ['click'], (eventClick) => {
            hideMessage();                       
            let view = Number(modalControls.hdnModalView.value);
            try {
                let player = null;
                switch (view) {
                    case ModalView.add:                        
                        player = new Game.players.Player();                     
                        break;
                    case ModalView.update: 
                        player = Game.players.get(modalControls.hdnPlayerId.value);
                        break;
                    default:
                        throw new TypeError(`Unknown view mode: ${view}`);
                        break;
                }            

                validateUpdate(player);

                player.name = modalControls.txtName.value;
                player.iconUrl = modalControls.txtIconUrl.value;
                player.showIcon = modalControls.chkShowIcon.checked;
                player.isAiPlayer = modalControls.chkIsAiPlayer.checked;
                player.difficulty = null;
                if (!Game.functions.isEmpty(modalControls.cboDifficulty.value)) {
                    player.difficulty = Number(modalControls.cboDifficulty.value);
                }
                
                let description = 'updated';
                if (view === ModalView.add) {
                    player = Game.players.create(player); 
                    description = 'added';
                }
                
                Game.functions.createPlayerRadioButtons(Game.players.get(), Game.players.getStarting());
                Game.functions.displayNotice(`${player.name} successfully ${description}!`, Notice.type.success);

                Modal.hide(modalElement);
                clearControls(clearDelay);                
            } catch (e) {
                // console.log(e);
                let message = e.message || e;
                showMessage(message);
            }
        });
    }

    exposed.getViewDescription = (view) => {
        let description = '';
        switch (view) {
            case ModalView.add:
                description = 'Add';
                break;
            case ModalView.update:
                description = 'Update';
                break;
        }
        return description;
    }

    let validateUpdate = (player = null) => {
        let problems = [];
        if (!Game.functions.isNull(player) 
            && !Game.players.isValidPlayer(player)) {
            problems.push('Error: Unable to determine the current player!');

        }
        if (Game.functions.isEmpty(modalControls.txtName.value)) {
            Game.element.tryFocus(modalControls.txtName);
            problems.push('Player name not specified')

        }                
        if (modalControls.chkShowIcon.checked 
                && Game.functions.isEmpty(modalControls.txtIconUrl.value)) {
            Game.element.tryFocus(modalControls.txtIconUrl);
            problems.push('Avatar url not specified');
        }
        if (modalControls.chkIsAiPlayer.checked 
                && Game.functions.isEmpty(modalControls.cboDifficulty.value)) {
            Game.element.tryFocus(modalControls.cboDifficulty);
            problems.push('CPU difficulty not specified');

        }          
        if (problems.length > 0) {
            let message = `
                Please correct the following:
                <ul>
                    ${problems.map((problem) => `<li>${problem}</li>`).join('\n')}
                </ul>
            `;
            throw new TypeError(message);
        }
    }

    let clearControls = (delay = 0) => {
        let clear = () => {
            let defaults = [
                {prop: 'value', default: null},
                {prop: 'checked', default: false},
            ];
            for (const prop in modalControls) {
                const control = modalControls[prop];
                if (Game.functions.isNull(control)) {
                    continue;
                } 
                for (const value of defaults) {
                    if (value.prop in control) {
                        control[value.prop] = value.default;
                    }
                }
            } 
            hideMessage();
            Game.element.resetScroll(Game.element.get('.modal-container'));
        }
        if (delay > 0) {
            setTimeout(()=> clear(), delay);
        } else {
            clear();
        }
        return modalControls;        
    }
    
    let getControls = () => {
        let elements = {
            txtName: '#name',
            chkShowIcon: '#showIcon',
            txtIconUrl: '#iconUrl',
            chkIsAiPlayer: '#isAiPlayer',
            cboDifficulty: '#aiDifficulty',
            hdnPlayerId: '#playerId',
            hdnModalView: '#modalView',
            btnEditPlayerCancel: '#btnEditPlayerCancel',
            btnEditPlayerSave: '#btnEditPlayerSave',
            divErrorMessage: '#errorMessage',
        };
        let modalControls = {};     
        for (const prop in elements) {
            let selector = elements[prop];
            modalControls[prop] = Game.element.get(selector);
        }
        return modalControls;
    } 
    
    let modalElement = null;
    let modalControls = null;
    let initComplete = false;
    let clearDelay = 200;    
}(Game.players.modal)); // http://programmingnotes.org/