// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  players.js
//    Description: Javascript that handles player related operations
// ============================================================================ 

/**
* NAMESPACE: Game.players
* USE: Handles gamestatus related operations.
*/

// Declare the namespace 
registerNamespace('Game.players');

// Declare the namespace members
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
    
    // Set class names and other shared data
    const settings = {
        classNameRadioImage: '.player-radio-image',
        classNameGamePieceImage: '.player-gamepiece-image',
        classNameGamePieceText: '.player-gamepiece-text',
        classNameRadioName: 'player-radio-name',
        classNameRadioNameDescription: 'player-radio-name-description',
        cleanClassName: function(str) {
            return str ? str.trim().replace('.', '') : '';
        },		
    };
    exposed.settings = settings;

    // Class to hold player information for the game
    exposed.Player = class {        
        constructor(id) {
            this._id = id;
            this.name  = null
            this.iconUrl  = false
            this.showIcon  = null
            this.isAiPlayer = false;
            this.difficulty = null
        }

        getId() {
            return this._id;
        }		
        
        getRadioButtonHtml() {
            let html = '';
            if (this.showIcon) {
                html += `
                    <span class="${settings.cleanClassName(settings.classNameRadioImage)}" style="background: 
                        url('${this.iconUrl}') no-repeat">
                    </span>
                `;
            }
            html += `<span class="${settings.cleanClassName(settings.classNameRadioName)}">
                ${this.name}
            `;
            html += `</span>`;
            if (this.isAiPlayer) {
                html += `<span class="${settings.cleanClassName(settings.classNameRadioNameDescription)}"> `;
                html += `- CPU`;                
                html += `: ${Game.ai.getDescription(this.difficulty)}`;
                html += `</span>`;
            }            
            return html;
        }	
        
        getGamePieceHtml() {
            let html = '';
            if (this.showIcon) {
                html += `
                    <span class="${settings.cleanClassName(settings.classNameGamePieceImage)}" style="background: 
                        url('${this.iconUrl}') no-repeat">
                    </span>
                `;
            } else {
                html += `<span class="${settings.cleanClassName(settings.classNameGamePieceText)}">
                    ${this.name[0].toUpperCase()}
                </span>`;
            }
            return html;
        }
    }

    /**
    * FUNCTION: create
    * USE: Creates a player object for the game and stores it in an internal array
    * @param name: The name of the player
    * @param iconUrl: The icon image url for the player
    * @param showIcon: Determines if the icon should be displayed
    * @return: The created player object.
    */		
    exposed.create = (options) => {
        let name = options.name;
        options.showIcon = options.showIcon != null ? options.showIcon : true;
        options.isAiPlayer = options.isAiPlayer != null ? options.isAiPlayer : false;

        if (!name || name.length < 1) {
            throw new Error('A name is required');
        }

        let id = getNextId();
        let newPlayer = new exposed.Player(id);
        let excluded = ['_id'];
        for (let prop in options) {
            let value = options[prop];
            if (newPlayer.hasOwnProperty(prop) 
                && typeof value !== 'undefined'
                && typeof value !== 'function'
                && excluded.indexOf(prop) < 0) {
                newPlayer[prop] = value;
            }
        }

        players.push(newPlayer);
        return newPlayer;
    }

    /**
    * FUNCTION: getIds
    * USE: Returns the players Ids for the players in the game. 
    *   If specific players are specified, only those are returned
    * @param playersToCheck: An array of players to get Ids for
    * @return: Player ids of the game.
    */		
    exposed.getIds = (playersToCheck = null) => {		
        if (!playersToCheck) {
            playersToCheck = exposed.get();
        }

        // Get the player Ids
        let playerIds = playersToCheck.map((player) => {
            return getPlayerId(player);
        });

        return playerIds;
    }	

    /**
    * FUNCTION: getId
    * USE: Returns the players Id for the players in the game. 
    * @param player: A player to get the Id for
    * @return: Player id of the game.
    */		
    exposed.getId = (player = null) => {		
        return getPlayerId(player);
    }    

    /**
    * FUNCTION: get
    * USE: Returns the players of the game. 
    *   If a player Id is specified, only that player is returned
    * @param id: Optional. Returns the player that matches the given Id
    * @return: Player(s) of the game.
    */	
    exposed.get = (id = null) => {
        let returnValue = players.slice();
        if (id) {
            id = getPlayerId(id);
            returnValue = players.find((x) => {
                return getPlayerId(x) == id;
            });
        }
        return returnValue;
    }
    
    /**
    * FUNCTION: setStarting
    * USE: Stores the starting player of a match
    * @param player: Player that should be the starting player of a match
    * @return: N/A.
    */		
    exposed.setStarting = (player) => {
        if (!player) {
            throw new Error('A starting player is required');
        }
        
        // Make sure the player exists
        player = exposed.get(player);

        // Make sure this is a player object
        if (!exposed.isValidPlayer(player)) {
            throw new Error('Unable to determine the starting player');
        }

        starting = player.getId();
    }

    /**
    * FUNCTION: getStarting
    * USE: Returns the Id of the starting player of a match
    * @return: The Id of the starting player.
    */		
    exposed.getStarting = () => {
        return starting;
    }

    /**
    * FUNCTION: isValidPlayer
    * USE: Determines if a value is a valid player object
    * @param player: A game player
    * @return: Returns if the given player is valid, false otherwise.
    */		
    exposed.isValidPlayer = (player) => {
        return player && (player instanceof exposed.Player);
    }

    /**
    * FUNCTION: clear
    * USE: Removes all the players of the game 
    * @return: N/A.
    */	
    exposed.clear = () => {
        starting = null;
        maxId = null;
        players.length = 0;
    }	

    /**
    * FUNCTION: load
    * USE: Removes the previous players and loads the 
    *    specified players into the game
    * @param loadPlayers: The players to be loaded into the game
    * @return: N/A.
    */	
    exposed.load = (gameSaveData, playersRequired = false) => {
        let loadPlayers = gameSaveData.players;
        if (playersRequired && (!loadPlayers || loadPlayers.length < 1)) {
            throw new Error('Unable to load. No players specified');
        }

        exposed.clear();
        if (loadPlayers) {
            for (const player of loadPlayers) {
                if (!exposed.isValidPlayer(player)) {
                    throw new Error(`Unable to load player of type ${typeof player}. Expected a game player object`);
                }
                players.push(player);
            }
        }

        if (!Game.functions.isNull(gameSaveData.playersMaxId)) {
            maxId = gameSaveData.playersMaxId;
        } 
        if (!maxId || maxId < players.length) {
            maxId = getNextId()-1;
        }                
    }

    /**
    * FUNCTION: remove
    * USE: Removes a player from the game 
    * @param player: Player that should be the starting player of a match
    * @return: True if player was removed, false otherwise.
    */	
    exposed.remove = (player) => {
        if (!player) {
            throw new Error('A player is required');
        }
        
        // Make sure the player exists
        player = exposed.get(player);

        // Make sure this is a player object
        if (!exposed.isValidPlayer(player)) {
            throw new Error('Unable to determine the player to remove');
        }

        let index = players.indexOf(player);
        if (index > -1) {
            players.splice(index, 1);
        }
        return index > -1;
    }	    

// -- Private data --
    let getPlayerId = (player) => {
        let id = player;
        if (typeof player == 'object' 
            && exposed.isValidPlayer(player)) {
            id = player.getId();
        }
        return id;
    }

    let getNextId = () => {
        maxId = maxId || 0;
        if (!maxId) {
            if (players && players.length > 0) {
                maxId = Game.functions.arrayMax(players, (player) => player.getId());
            }
        }
        maxId += 1;
        return maxId;
    }

    exposed.getMaxId = () => {
        return maxId;
    }

    let maxId = null;

    // Array that holds created players
    let players = [];

    let starting = null;
}(Game.players)); // http://programmingnotes.org/