// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  status.js
//    Description: Javascript that handles status related operations
// ============================================================================ 

/**
* NAMESPACE: Game.status
* USE: Handles gamestatus related operations.
*/

// Declare the namespace 
registerNamespace('Game.status');

// Declare the namespace members
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
    
    // Set class names and other shared data
    const settings = {	
    };
    exposed.settings = settings;
    
    /**
    * FUNCTION: checkWin
    * USE: Determines if the game has been won or not
    * @param lstPlayerMoves: An array of moves by a player
    * @return: Returns an object with a property ('winningComboFound')
    * 	that is set to true if the moves provided results in a winning
    *   state, and is set to false otherwise. The object also contains
    *   a property array ('winningCombo') that has the winning combination
    */	
    exposed.checkWin = (lstPlayerMoves) => {
        if (!lstPlayerMoves || !Array.isArray(lstPlayerMoves)) {
            throw new Error('An array of moves is required');
        }
        
        // Get the winning combos to check to see if a player has won
        let lstWinningCombos = exposed.getWinningCombos();
        if (!lstWinningCombos || !Array.isArray(lstWinningCombos)) {
            throw new Error('Unable to determine the winning moves');
        }
        
        // Determines if a player has won
        let winningComboFound = false;
        let winningCombo = null;
        
        // Go through the winning combos to check to see if a player has won
        for (let x = 0; x < lstWinningCombos.length; ++x) {
            // Get the current combo
            let lstCurrentCombo = lstWinningCombos[x];
            
            // Make sure this is a valid combo for the number of choices made
            if (lstCurrentCombo.length < 1 
                || lstPlayerMoves.length < lstCurrentCombo.length) {
                continue;
            }
            
            // Go through each element and see if the user selection matched a winning combo.
            let match = true;
            for (let y = 0; y < lstCurrentCombo.length; ++y) {
                let comboNumber = lstCurrentCombo[y];
                if (lstPlayerMoves.indexOf(comboNumber) === -1) {
                    // Exit loop. All elements have to match for this to be a winning combo
                    match = false;
                    break;
                }
            }
            
            // Check to see if we found a winning combo
            if (match) {
                winningComboFound = true;
                winningCombo = lstCurrentCombo;
                break;
            }
        }

        let result = {
            winningComboFound: winningComboFound,
            winningCombo: winningCombo
        };
        
        return result;
    }
    
    /**
    * FUNCTION: getWinningCombos
    * USE: Returns an 2d array of all the possible winning combinations a player can make
    * @return: Returns all the winning combinations for the game.
    */	
    exposed.getWinningCombos = () => {
        let winningCombos = [
            [1, 2, 3],          [1, 4, 7],      [1, 5, 9],
            [4, 5, 6],          [2, 5, 8],      [3, 5, 7],
            [7, 8, 9],          [3, 6, 9]
        /*Horizontal Wins*/     /*Vertical Wins*/       /*Diagonal Wins*/
        ];
        
        return winningCombos;
    }

    // Enum to hold the match result
    exposed.result = Object.freeze({none:0, winnerFound:1, tieGame:2});

    // Class to hold information for the current match
    exposed.Match = class {
        constructor(id, players, totalMovesPossible, startingPlayer) {
            this._id = id;
            this._players = Game.players.getIds(players);
            this._startingPlayer = startingPlayer;
            this._currentTurn = 0;
            this._winner = null;
            this._totalPlayers = players.length;
            this._totalMovesPossible = totalMovesPossible;
            this._totalMovesRemaining = totalMovesPossible;
            this._timeStart = new Date();
            this._timeEnd = null;
            this._matchStatus = null;

            // Create an array for each player that will store their moves
            this._playerMoves = {};
            this._playerInfo = {};
            players.forEach((player, index) => {
                let playerId = this._getPlayerId(player);
                this._playerMoves[playerId] = [];

                // Set the current turn index to the player 
                // that matches the starting player id
                if (playerId == startingPlayer) {
                    this._currentTurn = index;
                }

                // Set player info
                this._playerInfo[playerId] = {
                    isAiPlayer: player.isAiPlayer,
                    difficulty: player.difficulty
                };
            });
        }

        _setStatus(result) {
            this._matchStatus = result;
        }

        _getPlayerId(player) {
            return Game.players.getId(player);
        }	
        
        _getCurrentTurnIndex() {
            return this._currentTurn;
        }

        _peekNextTurnIndex(increment = 1) {
            return this._wrap(this._getCurrentTurnIndex() + increment, this.getTotalPlayers());
        }			

        _moveNextTurnIndex() {
            let nextTurn = this._peekNextTurnIndex();
            this._currentTurn = nextTurn
            return this._currentTurn;
        }
        
        // Circular indexing of an array
        _wrap(left, right) {
            return ((left % right) + right) % right;
        }

        _addMove(player, move) {
            let playerMoves = this.getMoves(player);
            playerMoves.push(move);
        }			
        
        _maybeSetTimeEnd(result, time) {
            if (this._timeEnd) {
                return;
            }

            // These are the conditions when to set the end time of the match
            let conditions = [exposed.result.winnerFound, exposed.result.tieGame];
            if (conditions.indexOf(result) > -1) {
                this._timeEnd = time;
            }
        }	

        _subtractMove() {
            this._totalMovesRemaining = Math.max(this.getTotalMovesRemaining() - 1, 0);
        }
        
        getStartingPlayer() {
            return this._startingPlayer;
        }

        getMoves(player = null) {
            let moves = this._playerMoves;
            if (player) {
                let id = this._getPlayerId(player);
                moves = moves[id];
            }
            return moves;            
        }

        getPlayerInfo(player = null) {
            let info = this._playerInfo;
            if (player) {
                let id = this._getPlayerId(player);
                info = info[id];
            }
            return info;
        }        

        getId() {
            return this._id;
        }
        
        getTotalPlayers() {
            return this._totalPlayers;
        }		

        getPlayers() {
            return this._players.slice();
        }
        
        getCurrentPlayer() {
            let players = this.getPlayers();
            let currentTurn = this._getCurrentTurnIndex();
            return players[currentTurn];
        }		

        peekNextPlayer(increment = 1) {
            let players = this.getPlayers();
            let nextTurn = this._peekNextTurnIndex(increment);
            return players[nextTurn];
        }

        moveNextPlayer() {
            let players = this.getPlayers();
            let nextTurn = this._moveNextTurnIndex();
            return players[nextTurn];
        }			

        getTotalMovesPossible() {
            return this._totalMovesPossible;
        }

        getTotalMovesRemaining() {
            return this._totalMovesRemaining;
        }

        getWinner() {
            return this._winner;
        }

        setWinner(player, winningCombo) {
            if (this._winner) {
                throw new Error('A winner has already been declared');
            }

            let result = {
                playerId: this._getPlayerId(player),
                winningCombo: winningCombo,
            };
            
            this._winner = result;
        }		

        addMove(move, player = null) {
            if (!player) {
                player = this.getCurrentPlayer();
            }
            
            this._addMove(player, move);
            this._subtractMove();
            this.moveNextPlayer();			
        }

        getDuration() {
            return exposed.getDuration(exposed.getDurationSeconds(this.getTimeStart(), this.getTimeEnd()));
        }

        getTimeStart() {
            return this._timeStart;
        }

        getTimeEnd() {
            return this._timeEnd;
        }

        getTimeStartString(format = '') {
            return exposed.getDateTimeString(this.getTimeStart(), format);
        }		

        getTimeEndString(format = '') {
            return exposed.getDateTimeString(this.getTimeEnd(), format);
        }

        getStatus() {
            return this._matchStatus;
        }

        checkWin(player = null) {
            let movesRemaining = this.getTotalMovesRemaining();
            let statusResult = exposed.result.none;
            let players = [];
            let now = new Date();

            if (player) {
                players.push(player);
            } else {
                players = this.getPlayers();
            }
            
            // Go through each player and check to see if a winner 
            // has been found, or if a tie game has occurred
            for (const player of players) {
                let moves = this.getMoves(player);
                if (moves.length > this.getTotalMovesPossible()) {
                    throw new Error(`Player Id #${this._getPlayerId(player)} has made more moves than allowed. (# Moves: ${moves.length})`);
                }
                //console.log(player);

                // Check to see if this player has any winning moves
                let checkWinResult = exposed.checkWin(moves);
                let hasWonGame = checkWinResult.winningComboFound;
                if (hasWonGame) {
                    if (!this.getWinner()) {
                        this.setWinner(player, checkWinResult.winningCombo);
                    }
                    statusResult = exposed.result.winnerFound;
                } else {				
                    if (movesRemaining < 1) {
                        statusResult = exposed.result.tieGame;
                    }
                }
                if (hasWonGame) {
                    break;
                }
            }

            this._maybeSetTimeEnd(statusResult, now);

            this._setStatus({
                result: statusResult
            });

            return this.getStatus();
        }
    }

    /**
    * FUNCTION: startNewMatch
    * USE: Creates a new match for the game
    * @param players: Players that will be in this match
    * @param totalMovesPossible: The total number of moves possible in the game
    * @param startingPlayer: The starting player of the match
    * @return: An object representing the newly created match.
    */	
    exposed.startNewMatch = (players, totalMovesPossible, startingPlayer) => {        
        if (players.length < 1) {            
            throw new TypeError(`Unable to start game. There are no players!`);
        }  
        let id = getNextId();    
        currentMatch = new exposed.Match(id, players, totalMovesPossible, startingPlayer);
        matches.push(currentMatch);
        return currentMatch;
    }

    /**
    * FUNCTION: getCurrentMatch
    * USE: Returns the most recent match created
    * @return: The most recent match.
    */	
    exposed.getCurrentMatch = () => {
        return currentMatch;
    }	

    /**
    * FUNCTION: getMatches
    * USE: Returns the matches that has been played in this current session
    * @return: Player matches.
    */		
    exposed.getMatches = () => {
        return matches;
    }

    /**
    * FUNCTION: hasStarted
    * USE: Determines if a match has started
    * @param match: Optional. Determines if the given match has started
    *   If no match is specified, the current match is used
    * @return: Returns if the given match has started, false otherwise.
    */		
    exposed.hasStarted = (match = null) => {
        if (!match) {
            match = exposed.getCurrentMatch();
        }
        return match != null && exposed.isValidMatch(match);
    }
    
    /**
    * FUNCTION: hasCompleted
    * USE: Determines if a match has ended
    * @param match: Optional. Determines if the given match has ended
    *   If no match is specified, the current match is used
    * @return: Returns if the given match has ended, false otherwise.
    */		
    exposed.hasCompleted = (match = null) => {
        if (!match) {
            match = exposed.getCurrentMatch();
        }
        return exposed.hasStarted(match) && (match.getWinner() || match.getTotalMovesRemaining() < 1)
    }

    /**
    * FUNCTION: isValidMatch
    * USE: Determines if a match is valid
    * @param match: A game match
    * @return: Returns if the given match is valid, false otherwise.
    */	
    exposed.isValidMatch = (match) => {
        return match && (match instanceof exposed.Match);
    }
    
    exposed.getResultDescription = (result) => {
        let description = '';

        switch (result) {
            case exposed.result.winnerFound:
                description = 'Winner';
                break;
            case exposed.result.tieGame:
                description = 'Tie';
                break;
        }
        
        return description.trim();
    }

    /**
    * FUNCTION: getDuration
    * USE: Takes a time in seconds and converts it to hours, mins, seconds
    * @param seconds: A time in seconds
    * @return: Returns an object which holds the converted time
    */		
    exposed.getDuration = (seconds) => { 
        let time = seconds;
        let hours = 0;
        let mins = 0;
        let secs = 0;

        if (seconds > 0) {
            hours = Math.floor(time / 3600);
            time = time % 3600;
            mins = Math.floor(time / 60);
            time = time % 60;
            secs = Number(time.toFixed(1));
        }

        return {
            hours: hours,
            mins: mins,
            secs: secs,
            totalSeconds: seconds,
            toString: function() {
                let duration = this;
                let string = '';
                if (duration.hours > 0) {
                    string += `${duration.hours} hour${duration.hours === 1 ? '' : 's'}`;
                }
                if (duration.mins > 0) {
                    string += ` ${duration.mins} min${duration.mins === 1 ? '' : 's'}`;
                }
                if (duration.secs > 0) {
                    string += ` ${duration.secs} sec${duration.secs === 1 ? '' : 's'}`;
                }
                return string.trim();
            },
        }
    }

    exposed.getDurationSeconds = (dateStart, dateEnd) => {
        let timeStart = dateStart ? dateStart.getTime() : 0;
        let timeEnd = dateEnd ? dateEnd.getTime() : 0;
        return Math.max((timeEnd - timeStart) / 1000, 0);
    }
    
    exposed.getDateTimeString = (date, format = '', nullDateString = 'N/A') => {
        let ret = nullDateString;
        if (date) {
            ret = '';
            format = format ? String(format).trim().toLocaleLowerCase() : format;

            if (!format || format.length < 1 || 'd' == format[0]) {
                ret = date.toLocaleDateString();
            } 

            if (!format || format.length < 1 || 't' == format[0]) {
                ret += ` ${date.toLocaleTimeString()}`;
            }
        }
        return ret.trim();
    }
    
    /**
    * FUNCTION: clear
    * USE: Removes all the matches of the game 
    * @return: N/A.
    */	
    exposed.clear = () => {
        currentMatch = null;
        maxId = null;
        matches.length = 0
    }

    /**
    * FUNCTION: load
    * USE: Removes the previous matches and loads the 
    *    specified matches into the game
    * @param loadMatches: The matches to be loaded into the game
    * @return: N/A.
    */	
    exposed.load = (gameSaveData, matchesRequired = false) => {
        let loadMatches = gameSaveData.matches
        if (matchesRequired && (!loadMatches || loadMatches.length < 1)) {
            throw new Error('Unable to load. No matches specified');
        }

        exposed.clear();
        let timeStart = null;
        if (loadMatches) {
            for (const match of loadMatches) {
                if (!exposed.isValidMatch(match)) {
                    throw new Error(`Unable to load match of type ${typeof match}. Expected a game match object`);
                }
                if (!timeStart || match.getTimeStart().getTime() > timeStart.getTime()) {
                    currentMatch = match;
                    timeStart = match.getTimeStart();
                }
                matches.push(match);
            }
        }

        if (!Game.functions.isNull(gameSaveData.matchesMaxId)) {
            maxId = gameSaveData.matchesMaxId;
        } 
        if (!maxId || maxId < matches.length) {
            maxId = getNextId()-1;
        }       
    }	
    
// -- Private data --
    // Array that holds match history
    let matches = [];

    let currentMatch = null;	

    let getNextId = () => {
        maxId = maxId || 0;
        if (!maxId) {
            if (matches && matches.length > 0) {
                maxId = Game.functions.arrayMax(matches, (match) => match.getId());
            }  
        }
        maxId += 1;
        return maxId;
    }

    exposed.getMaxId = () => {
        return maxId;
    }

    let maxId = null;    
}(Game.status)); // http://programmingnotes.org/