// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jul 7, 2020
//    Taken From: http://programmingnotes.org/
//    File:  Resources.js
//    Description: Javascript that loads Javascript and CSS files 
//       into a document in parallel, as fast as the browser will allow.
//       This allows to ensure proper execution order if you have dependencies 
//       between files.
//    Example:
//       // Load JavaScript and CSS files and wait for them to load
//       const dependencies = await Resources.load({
//           files: [	
//               { url: '/file1.js', type: 'js', wait: false },
//               { url: '/file2.js', type: 'js', wait: false },
//               { url: '/file3.css', type: 'css', wait: false },
//               { url: '/file4.css', type: 'css', wait: false },
//               { url: 'https://code.jquery.com/jquery-3.5.1.js', type: 'js', wait: true },
//               { url: 'https://ajax.googleapis.com/ajax/libs/jquery/3.5.1/jquery.min.js', type: 'js', wait: false },
//            ],
//            logResults: true,
//            rejectOnError: false,
//            timeout: null
//        });	
// ============================================================================
/**
* NAMESPACE: Resources
* USE: Handles loading Javascript and CSS files in parallel.
*/
var Resources = Resources || {};
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
    
    /**
    * FUNCTION: load
    * USE: Loads Css and/or Javascript files into a document in 
    *    parallel, depending on the options 
    * @param options: An object of file loading options. 
    *   Its made up of the following properties:
    *   {
    *       files: An object array of File info of files to load (See below).
    *       logResults: Optional. Boolean that indicates if the load 
    *          status should be displayed to console.log().
    *          Default is false.
    *       rejectOnError: Optional. Boolean that indicates if the load 
    *          operation should reject if a loading error occurs for 
    *          any file in the group. 
    *          Default is false.
    *       timeout: Optional. Integer that indicates how long to wait
    *          (in milliseconds) for the file group to load before timeout occurs. 
    *          Default is null.  
    *   }
    * 
    *   ** File info: This is made up of the following properties
    *   {
    *       url: The source url to load.
    *       type: Optional. String indicating the file type to load. 
    *          Options are 'css' or 'js'.
    *          Default type is determined by the file extension.
    *       wait: Optional. Boolean that indicates if this file should be
    *          completely loaded before loading subsequent files in the group.          
    *               If set to true, other urls are not loaded until this one 
    *                   finishes loading. After loading this file, other files
    *                   continue loading in parallel.
    *               If set to false, no wait occurs and files load in parallel.
    *          This makes it so dependent files dont get loaded until the 
    *          required files are loaded first.
    *          Files with this property set to true are loaded before 
    *          any other file in the group.
    *          Default is false.
    *   }
    * @return: A promise that completes when all resources are loaded, which
    *          contains an array of each files loading status.
    */     
    exposed.load = (options) => { 
        return new Promise(async (resolve, reject) => {
            try {
                // Load files and wait for the results
                options = verifyOptions(options);
                let results = await loadFiles(options);
                if (options.logResults) {
                    console.log('*** Resources.js Load Status ***')
                    let notes = results.map((x) => x.notes);
                    console.log(notes.join('\n'));
                }  
                resolve(results); 
            } catch (e) {
                reject(e);
            }
        });
    }

    /**
    * FUNCTION: loadJs
    * USE: Loads a Javascript file into a document 
    * @param url: The javascript url to load.
    * @return: A promise that completes when the url is loaded.
    */     
    exposed.loadJs = (url) => {
        return new Promise((resolve, reject) => {
            // Load file and attach it to the document
            let status = createFileStatus(url);
            try {
                if (fileExists(url)) {
                    status.loaded = true;
                    status.notes = `${url} already loaded`;
                    resolve(status);
                    return;
                }
                let script = document.createElement("script");
                script.src = url;
                script.type = "text/javascript"
                script.id = getId(url);
                script.onload = script.onreadystatechange = function() {
                    if ((!this.readyState 
                        || this.readyState == "loaded" 
                        || this.readyState == "complete") ) {
                            status.loaded = true;
                            status.notes = `${url} loaded`;
                            resolve(status);
                    }	
                }
                script.onerror = () => {
                    removeFile(url);
                    status.notes = `*** ${url} failed to load ***`;
                    reject(status);
                }
                document.getElementsByTagName("head")[0].appendChild(script);
                addFile(url);
            } catch (e) {
                let message = e.message ? e.message : e;
                status.notes = `*** ${url} failed to load. Reason: ${message} ***`;
                reject(status);
            }
        });
    };	

    /**
    * FUNCTION: loadCss
    * USE: Loads a Css file into a document 
    * @param url: The css url to load.
    * @return: A promise that completes when the url is loaded.
    */       
    exposed.loadCss = (url) => {
        return new Promise((resolve, reject) => {
            // Load file and attach it to the document
            let status = createFileStatus(url);
            try {	
                if (fileExists(url)) {
                    status.loaded = true;
                    status.notes = `${url} already loaded`;
                    resolve(status);
                    return;
                }
                let link = document.createElement('link');
                link.type = 'text/css';
                link.rel = 'stylesheet';
                link.onload = () => { 
                    status.loaded = true;
                    status.notes = `${url} loaded`;
                    resolve(status); 
                }
                link.onerror = () => {
                    removeFile(url);
                    status.notes = `*** ${url} failed to load ***`;
                    reject(status);
                } 
                link.href = url;
                link.id = getId(url);
                document.getElementsByTagName("head")[0].appendChild(link);  
                addFile(url);              
            } catch (e) {
                let message = e.message ? e.message : e;
                status.notes = `*** ${url} failed to load. Reason: ${message} ***`;
                reject(status);
            }
        });
    };	

    /**
    * FUNCTION: getFilename
    * USE: Returns the filename of a url
    * @param url: The url.
    * @return: The filename of a url.
    */    
    exposed.getFilename = (url) => {
        let filename = '';
        if (url && url.length > 0) {
            filename = url.split('\\').pop().split('/').pop();
            // Remove any querystring
            if (filename.indexOf('?') > -1) {
                filename = filename.substr(0, filename.indexOf('?'));
            }
        }
        return filename;
    }
    
    /**
    * FUNCTION: getExtension
    * USE: Returns the file extension of the filename of a url
    * @param url: The url.
    * @return: The file extension of a url.
    */     
    exposed.getExtension = (url) => {
        let filename = exposed.getFilename(url);
        let ext = filename.split('.').pop();
        return (ext === filename) ? '' : ext;
    }  
    
// -- Private data --
    let loadFiles = (options) => {
        return new Promise(async (resolve, reject) => {            
            try {
                if (!options.files) {
                    throw new Error('There are no files specified to load.');
                }
                let files = verifyArray(options.files);

                // Order the files putting the critical ones first
                prioritizeFiles(files);

                // Load files to the document
                let results = [];
                let promises = [];
                for (let file of files) {
                    file = verifyFile(file);
                    let promise = getCallback(file).call(this, file.url);
                    if (file.wait) {
                        let value = await resolveByCompletion([promise], options.rejectOnError, options.timeout);
                        results.push(value[0]);
                    } else {
                        promises.push(promise);
                    }
                }

                // Get the results
                let completed = await resolveByCompletion(promises, options.rejectOnError, options.timeout);
                results = [...results, ...completed];
                resolve(results);
            } catch (e) {
                let message = e.message ? e.message : e.notes ? e.notes : e;
                reject(new Error(`Unable to load resources. Reason: ${message}`));
            }
        });
    }

    let prioritizeFiles = (files) => {
        files.sort((x, y) => {
            let conditionX = x.wait || false;
            let conditionY = y.wait || false;
            return conditionY - conditionX;
        }); 
        return files;       
    }

    let getCallback = (file) => {
        let types = {
            css: exposed.loadCss,
            js: exposed.loadJs
        }
        let callback = types[file.type.toLocaleLowerCase()];
        if (!callback) {
            throw new Error(`Unknown file type/file extension (${file.type}) for the following url: ${file.url}`);
        }
        return callback;
    }

    let getId = (str) => {
        return `${str}_added_resource`
    }

    let fileExists = (url) => {
        return loadedResources.has(url) 
            || (document.querySelector(`script[src='${url}']`) != null);
    }

    let addFile = (url) => {
        loadedResources.set(url, true);
    }

    let removeFile = (url) => {
        return loadedResources.delete(url);
    }  
    
    let createFileStatus = (url) => {
        let status = {
            url: url,
            loaded: false,
            notes: '',
        }
        return status;
    }

    let verifyOptions = (options) => {
        if (!options) {
            throw new Error('There are no options specified');
        }
        if (typeof options !== 'object' || Array.isArray(options)) {
            let files = options;
            options = {};
            options.files = files;
        }
        options.rejectOnError = options.rejectOnError || false;
        options.timeout = options.timeout|| null;
        return options;
    }        

    let verifyArray = (arry) => {
        if (!arry) {
            arry = [];
        }
        if (!Array.isArray(arry)) {
            arry = [arry];
        }
        return arry;
    }

    let verifyFile = (file) => {
        if (typeof file !== 'object') {
            let url = file;
            file = {};
            file.url = url;
        }
        file.url = file.url || '';
        file.wait = file.wait || false;
        file.type = file.type || exposed.getExtension(file.url);
        return file
    }
    
    /**
    * FUNCTION: resolveByCompletion
    * USE: Returns a promise that will resolve when all of the input's 
    *    promises have resolved, in order of completion.
    * @param promises: An array of promises.
    * @param rejectOnError: Optional. Boolean that indicates if the operation 
    *    should reject if any of the input promises reject or throw an error.
    *    If set to true, the operation is rejected if any of the input 
    *        promises reject or throw an error.
    *    If set to false, the promises rejected value is added to the result list.	
    * @param timeout: Optional. Integer that indicates how long to wait
    *    (in milliseconds) for the promise group to complete.
    * @return: A promise that will contain the input promises results on completion.
    */
   let resolveByCompletion = (promises, rejectOnError = true, timeout = null) => {
        return new Promise(async (resolve, reject) => {
            try {
                let results = []
                let promiseMap = new Map();
                Array.prototype.forEach.call(promises, (promise, index) => {
                    let promiseResult = { index: index, value: null, error: null };
                    let mapValue = null;
                    if (promise instanceof Promise) {
                        mapValue = promise
                        .then(value => {promiseResult.value = value; return promiseResult})
                        .catch(error => {promiseResult.error = error; return promiseResult})
                    } else {
                        mapValue = promiseResult;
                        promiseResult.value = promise;						
                    }
                    promiseMap.set(index, mapValue);
                });

                let start = Date.now();
                let isTimedOut = () => {
                    let result = false;
                    if (timeout) {
                        let elapsed = (Date.now() - start);
                        result = elapsed >= timeout;
                    }
                    return result;
                }

                while (promiseMap.size > 0) {
                    let promiseResult = await Promise.race(promiseMap.values());
                    if (!promiseMap.delete(promiseResult.index)) {
                        throw new Error('Error occurred processing values');             
                    }                    
                    if (promiseResult.error) {
                        if (rejectOnError) {
                            reject(promiseResult.error);
                            return;
                        }
                        results.push(promiseResult.error);
                    } else {
                        results.push(promiseResult.value);
                    }                    
                    if (isTimedOut()) {
                        throw new Error(`Timeout of ${timeout}ms expired`);
                    } 
                }
                resolve(results);
            } catch (e) {
                reject(e);
            }
        });
    }

    let loadedResources = new Map();
    
    (function (factory) {
        if (typeof define === 'function' && define.amd) {
            define([], factory);
        } else if (typeof exports === 'object') {
            module.exports = factory();
        }
    }(function() { return namespace; }));     
}(Resources)); // http://programmingnotes.org/