// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 12, 2020
//    Taken From: http://programmingnotes.org/
//    File:  Notice.js
//    Description: Javascript that displays a Notice to the screen
// ============================================================================
/**
* NAMESPACE: Notice
* USE: Displays a Notice to the screen.
*/
var Notice = Notice || {};
(function(namespace) {
    'use strict';

// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;

    // Set class names and other shared data
    const settings = {
        classNameNotice: '.notice',
        classNameSuccess: '',
        classNameWarning: '.warning',
        classNameError: '.error',
        classNameShow: '.show',
        classNameHide: '.hide',
        cleanClassName: function(str) {
            return str ? str.trim().replace('.', '') : '';
        },
    };
    exposed.settings = settings;

    // Notice type enum
    const NoticeType = Object.freeze({success:1, warning:2, error:3});
    exposed.type = NoticeType;

    /**
    * FUNCTION: display
    * USE: Displays a message to the page with the given text.
    * @param text: Text to be displayed to the page.
    * @param type: The message type. The type determines the style of the notice.
    * @param duration: The length the notice is displayed to the page (in milliseconds).
    *   Set value to null to display the notice indefinitely.
    * @return: N/A.
    */
    exposed.display = (element, text, type = Notice.type.success, duration = 3000) => {
        // Make sure the notice element exists
        if (!element) {
            throw new Error('Notice element does not exist');
        }

        let id = getId(element);
        // Remove any previously existing notice duration delay
        if (noticeClearTimeout[id]) {
            clearTimeout(noticeClearTimeout[id]);
        }

        // Style the notice depending on the notice type
        applyStyleType(element, type);

        // Set the notice text
        element.innerHTML = text;

        // Fade in the notice to make it visible
        fadeIn(element);

        // Set the notice display duration
        if (duration) {
            noticeClearTimeout[id] = setTimeout(() => {
                // Fade out the notice to make it invisible
                fadeOut(element);
                delete noticeClearTimeout[id];
            }, duration);
        }
    }
// -- Private data --
    // Holds the notice duration timeout delay
    let noticeClearTimeout = {};

    let applyStyleType = (element, type) => {
        // Define the css classes for the different notice types
        let styles = getStyles();

        // Remove previous styles for the notice and set the new one
        for (const style of styles) {
            if (style.class.length < 1) {
                continue;
            }
            // Check if the class is applied
            let classExists =  hasClass(element, style.class);

            // Add the selected class
            if (type && type === style.type) {
                if (!classExists) {
                    addClass(element, style.class);
                }
            } else if (classExists) {
                // Remove class if exists
                removeClass(element, style.class);
            }
        }
    }

    let getId = (element) => {
        if (!element.id || element.id.trim().length < 1) {
            element.id = `notice_${randomFromTo(1271991, 7281987)}`;
        }
        return element.id;
    }

    let randomFromTo = (from, to) => {
        return Math.floor(Math.random() * (to - from + 1) + from);
    }

    let getStyles = () => {
        let styles = [
            {type: NoticeType.success, class: settings.classNameSuccess},
            {type: NoticeType.warning, class: settings.classNameWarning},
            {type: NoticeType.error, class: settings.classNameError}
        ];
        return styles;
    }

    let fadeIn = (element) => {
        removeClass(element, settings.classNameHide);
        addClass(element, settings.classNameShow);
    }

    let fadeOut = (element) => {
        removeClass(element, settings.classNameShow);
        addClass(element, settings.classNameHide);
    }

    let addClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && !hasClass(element, cssClass)) {
            element.classList.add(cssClass)
            modified = true;
        }
        return modified;
    }

    let removeClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && hasClass(element, cssClass)) {
            element.classList.remove(cssClass);
            modified = true;
        }
        return modified;
    }

    let hasClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);
        return element.classList.contains(cssClass);
    }

    (function (factory) {
        if (typeof define === 'function' && define.amd) {
            define([], factory);
        } else if (typeof exports === 'object') {
            module.exports = factory();
        }
    }(function() { return namespace; }));
}(Notice)); // http://programmingnotes.org/